#ifndef _DYNAMIXEL_MASTER_H
#define _DYNAMIXEL_MASTER_H

#ifdef __cplusplus
extern "C" {
#endif

#include "comm.h"
#include "dynamixel.h"
#include "dynamixel2.h"

#ifndef MAX_DYN_MASTER_TX_BUFFER_LEN
  #error "Please, specify the maximum TX buffer length with the MAX_DYN_MASTER_TX_BUFFER_LEN macro"
#endif

#ifndef MAX_DYN_MASTER_RX_BUFFER_LEN
  #error "Please, specify the maximum RX buffer length with the MAX_DYN_MASTER_TX_BUFFER_LEN macro"
#endif

/**
 * \brief
 *
 */
typedef struct
{
  /**
   * \brief
   *
   */
  TComm *comm_dev;
  /**
   * \brief
   *
   */
  TDynVersion version;
  /**
   * \brief
   *
   */
  unsigned short int op_length;
  /**
   * \brief
   *
   */
  unsigned char tx_buffer[MAX_DYN_MASTER_TX_BUFFER_LEN];
  /**
   * \brief
   *
   */
  unsigned char rx_buffer[MAX_DYN_MASTER_RX_BUFFER_LEN];
  /**
   * \brief
   *
   */
  return_level_t return_level;
  /**
   * \brief
   *
   */
  unsigned char packet_ready;
  /**
   * \brief
   *
   */
  unsigned char received_bytes;
  /**
   * \brief
   *
   */
  void (*set_tx_mode)(void);
  /**
   * \brief
   *
   */
  void (*set_rx_mode)(void);
  /**
   * \brief
   *
   */
  unsigned short int rx_timeout_ms;
  /**
   * \brief
   *
   */
  unsigned char rx_no_answer;
  /**
   * \brief
   *
   */
  unsigned char rx_num_packets;
}TDynamixelMaster;

/* public functions */
/**
 * \brief
 *
 */
void dyn_master_init(TDynamixelMaster *master,TComm *dev,TDynVersion version);
/**
 * \brief
 *
 */
static inline void dyn_master_set_rx_timeout(TDynamixelMaster *master,unsigned short int timeout_ms)
{
  if(master!=0x00000000)
    master->rx_timeout_ms=timeout_ms;
}
/**
 * \brief
 *
 */
static inline void dyn_master_set_return_level(TDynamixelMaster *master,return_level_t level)
{
  if(master!=0x00000000)
    master->return_level=level;
}
/**
 * \brief
 *
 */
static inline return_level_t dyn_master_get_return_level(TDynamixelMaster *master)
{
  if(master!=0x00000000)
    return master->return_level;
  else 
    return return_all; 
}
/**
 * \brief
 *
 */
void dyn_master_scan(TDynamixelMaster *master,unsigned char *num,unsigned char *ids);
/**
 * \brief
 *
 */
unsigned char dyn_master_ping(TDynamixelMaster *master,unsigned char id);
/**
 * \brief
 *
 */
unsigned char dyn_master_read_byte(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned char *data);
/**
 * \brief
 *
 */
unsigned char dyn_master_read_word(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int *data);
/**
 * \brief
 *
 */
unsigned char dyn_master_read_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data);
/**
 * \brief
 *
 */
unsigned char dyn_master_write_byte(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned char data);
/**
 * \brief
 *
 */
unsigned char dyn_master_write_word(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned short int data);
/**
 * \brief
 *
 */
unsigned char dyn_master_write_table(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned short int length,unsigned char *data);
/**
 * \brief
 *
 */
unsigned char dyn_master_reg_write(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned short int length,unsigned char *data);
/**
 * \brief
 *
 */
unsigned char dyn_master_action(TDynamixelMaster *master);
/**
 * \brief
 *
 */
unsigned char dyn_master_sync_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[]);
/**
 * \brief
 *
 */
unsigned char dyn_master_sync_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[]);
/**
 * \brief
 *
 */
unsigned char dyn_master_bulk_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[]);
/**
 * \brief
 *
 */
unsigned char dyn_master_bulk_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[]);
/**
 * \brief
 *
 */
unsigned char dyn_master_relay(TDynamixelMaster *master,unsigned char *inst_pkt,unsigned char *status_pkt);

#ifdef __cplusplus
}
#endif

#endif
