#ifndef _DYNAMIXEL_SLAVE_H
#define _DYNAMIXEL_SLAVE_H

#ifdef __cplusplus
extern "C" {
#endif

#include "comm.h"
#include "dynamixel.h"
#include "dynamixel2.h"
#include "scheduler.h"
#include "dynamixel_slave_device.h"

#ifndef MAX_DYN_SLAVE_TX_BUFFER_LEN
  #error "Please, specify the maximum TX buffer length with the MAX_DYN_SLAVE_TX_BUFFER_LEN macro"
#endif

#ifndef MAX_DYN_SLAVE_RX_BUFFER_LEN
  #error "Please, specify the maximum RX buffer length with the MAX_DYN_SLAVE_RX_BUFFER_LEN macro"
#endif

#ifndef MAX_NUM_SLAVE_DEVICES
  #error "Please, specify the maximum number of slave devices with the MAX_NUM_SLAVE_DEVICES macro"
#endif

/**
 * \brief
 *
 */
typedef struct TDynamixelSlave
{
  /**
   * \brief
   *
   */
  TScheduler *scheduler;
  /**
   * \brief
   *
   */
  sched_channel_t sch_channel;
  /**
   * \brief
   *
   */
  TComm *comm_dev;
  /**
   * \brief
   *
   */
  TDynVersion version;
  /**
   * \brief
   *
   */
  unsigned char packet_ready;
  /**
   * \brief
   *
   */
  unsigned char received_bytes;
  /**
   * \brief
   *
   */
  unsigned short int rx_timeout_ms;
  /**
   * \brief
   *
   */
  unsigned short int op_length;
  /**
   * \brief
   *
   */
  unsigned char tx_buffer[MAX_DYN_SLAVE_TX_BUFFER_LEN];
  /**
   * \brief
   *
   */
  unsigned char rx_buffer[MAX_DYN_SLAVE_RX_BUFFER_LEN];
  /**
   * \brief
   *
   */
  unsigned char num_slave_devices;
  /**
   * \brief
   *
   */
  TDynamixelSlaveDevice *slave_devices[MAX_NUM_SLAVE_DEVICES];
  /**
   * \brief
   *
   */
  void (*set_tx_mode)(void);
  /**
   * \brief
   *
   */
  void (*set_rx_mode)(void);
  /**
   * \brief
   *
   */
  void (*set_baudrate)(TComm *comm_dev,unsigned int baudrate);
  /**
   * \brief
   *
   */
  unsigned char (*on_relay)(TDynVersion version,unsigned char *inst_pkt,unsigned char *status_pkt);
}TDynamixelSlave;

/* public functions */
/**
 * \brief
 *
 */
unsigned char dyn_slave_init(TDynamixelSlave *slave,TComm *dev,TScheduler *scheduler,sched_channel_t ch,TDynVersion version);
/**
 * \brief
 *
 */
unsigned char dyn_slave_add_device(TDynamixelSlave *slave, TDynamixelSlaveDevice *device);
/**
 * \brief
 *
 */
static inline void dyn_slave_set_baudrate(TDynamixelSlave *slave,unsigned int baudrate)
{
  if(slave!=0x00000000)
    slave->set_baudrate(slave->comm_dev,baudrate);
}
/**
 * \brief
 *
 */
static inline void dyn_slave_set_rx_timeout(TDynamixelSlave *slave,unsigned short int timeout_ms)
{
  if(slave!=0x00000000)
    slave->rx_timeout_ms=timeout_ms;
}
/**
 * \brief
 *
 */
static inline TDynVersion dyn_slave_get_version(TDynamixelSlave *slave)
{
  if(slave!=0x00000000)
    return slave->version;
  else
    return DYN_VER1;
}
/**
 * \brief
 *
 */
static inline void dyn_slave_start(TDynamixelSlave *slave)
{
  scheduler_enable_channel(slave->scheduler,slave->sch_channel);
}
/**
 * \brief
 *
 */
static inline void dyn_slave_stop(TDynamixelSlave *slave)
{
  scheduler_disable_channel(slave->scheduler,slave->sch_channel);
}

#ifdef __cplusplus
}
#endif

#endif
