#include "dynamixel.h"
#include "dynamixel2.h"

// provate functions
void dyn_set_checksum(unsigned char *packet)
{
  unsigned char checksum=0x00;
  unsigned char i;

  checksum+=packet[DYN_ID_OFF];
  checksum+=packet[DYN_LENGTH_OFF];
  checksum+=packet[DYN_INST_OFF];
  for(i=0;i<packet[DYN_LENGTH_OFF]-2;i++)
    checksum+=packet[DYN_DATA_OFF+i];
  packet[DYN_DATA_OFF+i]=~checksum;
}

unsigned char dyn_check_checksum(unsigned char *packet)
{
  unsigned char checksum=0x00;
  unsigned char i;

  checksum+=packet[DYN_ID_OFF];
  checksum+=packet[DYN_LENGTH_OFF];
  checksum+=packet[DYN_INST_OFF];
  for(i=0;i<packet[DYN_LENGTH_OFF]-2;i++)
    checksum+=packet[DYN_DATA_OFF + i];
  checksum+=packet[DYN_DATA_OFF+i];

  return checksum;
}

// public functions
void dyn_copy_packet(unsigned char *source, unsigned char *destination)
{
  unsigned char i,length;

  length=source[DYN_LENGTH_OFF]+4;
  for(i=0;i<length;i++)
    destination[i]=source[i];
}

unsigned char dyn_convert_v2_inst_packet(unsigned char *source, unsigned char *destination)
{
  unsigned char i;

  if(dyn2_get_length(source)>255)
    return 0x00;
  else
  {
    destination[DYN_HEADER_OFF]=0xFF;
    destination[DYN_HEADER_OFF+1]=0xFF;
    destination[DYN_ID_OFF]=source[DYN2_ID_OFF];
    destination[DYN_LENGTH_OFF]=source[DYN2_LENGTH_OFF];
    destination[DYN_INST_OFF]=source[DYN2_INST_OFF];
    for(i=0;i<destination[DYN_LENGTH_OFF];i++)
      destination[DYN_DATA_OFF+i]=source[DYN2_DATA_OFF+i];
    dyn_set_checksum(destination);
    return 0x01;
  }
}

/* ping instruction*/
void dyn_init_ping_packet(unsigned char *packet,unsigned char id)
{
  packet[DYN_HEADER_OFF]=0xFF;
  packet[DYN_HEADER_OFF+1]=0xFF;
  packet[DYN_ID_OFF]=id;
  packet[DYN_LENGTH_OFF]=0x02;
  packet[DYN_INST_OFF]=DYN_PING;
  dyn_set_checksum(packet);
}

/* read instruction */
void dyn_init_read_packet(unsigned char *packet,unsigned char id,unsigned char address,unsigned char length)
{
  packet[DYN_HEADER_OFF]=0xFF;
  packet[DYN_HEADER_OFF+1]=0xFF;
  packet[DYN_ID_OFF]=id;
  packet[DYN_LENGTH_OFF]=0x04;
  packet[DYN_INST_OFF]=DYN_READ;
  packet[DYN_DATA_OFF]=address;
  packet[DYN_DATA_OFF+1]=length;
  dyn_set_checksum(packet);
}

/* write instruction */
void dyn_init_write_packet(unsigned char *packet,unsigned char id,unsigned char address,unsigned char length,unsigned char *data)
{
  unsigned char i;

  packet[DYN_HEADER_OFF]=0xFF;
  packet[DYN_HEADER_OFF+1]=0xFF;
  packet[DYN_ID_OFF]=id;
  packet[DYN_LENGTH_OFF]=0x03+length;
  packet[DYN_INST_OFF]=DYN_WRITE;
  packet[DYN_DATA_OFF]=address;
  for(i=0;i<length;i++)
    packet[DYN_DATA_OFF+i+1]=data[i];
  dyn_set_checksum(packet);
}

unsigned char dyn_get_write_data(unsigned char *packet,unsigned char *data)
{
  unsigned char i;

  for(i=0;i<packet[DYN_LENGTH_OFF]-3;i++)
    data[i]=packet[DYN_DATA_OFF+i+1];

  return packet[DYN_LENGTH_OFF]-3;
}

/* registered write instruction */
void dyn_init_reg_write_packet(unsigned char *packet,unsigned char id,unsigned char address,unsigned char length,unsigned char *data)
{
  unsigned char i;

  packet[DYN_HEADER_OFF]=0xFF;
  packet[DYN_HEADER_OFF+1]=0xFF;
  packet[DYN_ID_OFF]=id;
  packet[DYN_LENGTH_OFF]=0x03+length;
  packet[DYN_INST_OFF]=DYN_REG_WRITE;
  packet[DYN_DATA_OFF]=address;
  for(i=0;i<length;i++)
    packet[DYN_DATA_OFF+i+1]=data[i];
  dyn_set_checksum(packet);
}

unsigned char dyn_get_reg_write_data(unsigned char *packet,unsigned char *data)
{
  unsigned char i;

  for(i=0;i<packet[DYN_LENGTH_OFF]-3;i++)
    data[i]=packet[DYN_DATA_OFF+i+1];

  return packet[DYN_LENGTH_OFF]-3;
}

/* action instruction */
void dyn_init_action_packet(unsigned char *packet)
{
  packet[DYN_HEADER_OFF]=0xFF;
  packet[DYN_HEADER_OFF+1]=0xFF;
  packet[DYN_ID_OFF]=0xFE;
  packet[DYN_LENGTH_OFF]=0x02;
  packet[DYN_INST_OFF]=DYN_ACTION;
  dyn_set_checksum(packet);
}

/* reset instruction */
void dyn_init_reset_packet(unsigned char *packet,unsigned char id)
{
  packet[DYN_HEADER_OFF]=0xFF;
  packet[DYN_HEADER_OFF+1]=0xFF;
  packet[DYN_ID_OFF]=id;
  packet[DYN_LENGTH_OFF]=0x02;
  packet[DYN_INST_OFF]=DYN_RESET;
  dyn_set_checksum(packet);
}

/* sync write instruction */
void dyn_init_sync_write_packet(unsigned char *packet,unsigned char num_servos,unsigned char *servo_ids,unsigned char address,unsigned char length,unsigned char * const data[])
{
  unsigned char i,j;

  packet[DYN_HEADER_OFF]=0xFF;
  packet[DYN_HEADER_OFF+1]=0xFF;
  packet[DYN_ID_OFF]=0xFE;
  packet[DYN_LENGTH_OFF]=0x04+(length+1)*num_servos;
  packet[DYN_INST_OFF]=DYN_SYNC_WRITE;
  packet[DYN_DATA_OFF]=address;
  packet[DYN_DATA_OFF+1]=length;
  for(i=0;i<num_servos;i++)
  {
    packet[DYN_DATA_OFF+2+(length+1)*i]=servo_ids[i];
    for(j=0;j<length;j++)
      packet[DYN_DATA_OFF+2+(length+1)*i+j+1]=data[i][j];
  }
  dyn_set_checksum(packet);
}

unsigned char dyn_sync_write_id_present(unsigned char *packet,unsigned char id,unsigned char *address,unsigned char *length,unsigned char *data)
{
  unsigned char num_servos,i,j;

  num_servos=(packet[DYN_LENGTH_OFF]-0x04)/(packet[DYN_DATA_OFF+1]+1);
  for(i=0;i<num_servos;i++)
    if(packet[DYN_DATA_OFF+2+(packet[DYN_DATA_OFF+1]+1)*i]==id)
    {
      (*address)=packet[DYN_DATA_OFF];
      (*length)=packet[DYN_DATA_OFF+1];
      for(j=0;j<packet[DYN_DATA_OFF+1];j++)
        data[j]=packet[DYN_DATA_OFF+2+(packet[DYN_DATA_OFF+1]+1)*i+j+1];
      return 0x01;
    }
  return 0x00;
}

/* bulk read instruction */
void dyn_init_bulk_read_packet(unsigned char *packet,unsigned char num_servos,unsigned char *servo_ids,unsigned char *address,unsigned char *length)
{
  unsigned char i;

  packet[DYN_HEADER_OFF]=0xFF;
  packet[DYN_HEADER_OFF+1]=0xFF;
  packet[DYN_ID_OFF]=0xFE;
  packet[DYN_LENGTH_OFF]=0x03;
  packet[DYN_INST_OFF]=DYN_BULK_READ;
  packet[DYN_DATA_OFF]=0x00;
  for(i=0;i<num_servos;i++)
  {
    packet[DYN_DATA_OFF+1+i*3]=length[i];
    packet[DYN_DATA_OFF+1+i*3+1]=servo_ids[i];
    packet[DYN_DATA_OFF+1+i*3+2]=address[i];
    packet[DYN_LENGTH_OFF]+=0x03;
  }
  dyn_set_checksum(packet);
}

unsigned char dyn_bulk_read_id_present(unsigned char *packet,unsigned char id,unsigned char *address,unsigned char *length)
{
  unsigned char num_servos,i,prev_id=0x00;

  num_servos=(packet[DYN_LENGTH_OFF]-0x03)/3;
  for(i=0;i<num_servos;i++)
  {
    if(packet[DYN_DATA_OFF+1+i*3+1]==id)
    {
      (*address)=packet[DYN_DATA_OFF+1+i*3+2];
      (*length)=packet[DYN_DATA_OFF+1+i*3];
      return prev_id;
    }
    else
      prev_id=packet[DYN_DATA_OFF+1+i*3+1];
  }
  return 0xFF;
}

unsigned char dyn_convert_v2_status_packet(unsigned char *source, unsigned char *destination)
{
  unsigned char i;

  if(dyn2_get_length(source)>255)
    return 0x00;
  else
  {
    destination[DYN_HEADER_OFF]=0xFF;
    destination[DYN_HEADER_OFF+1]=0xFF;
    destination[DYN_ID_OFF]=source[DYN2_ID_OFF];
    destination[DYN_LENGTH_OFF]=source[DYN2_LENGTH_OFF];
    destination[DYN_ERROR_OFF]=source[DYN2_ERROR_OFF];
    for(i=0;i<destination[DYN_LENGTH_OFF];i++)
      destination[DYN_DATA_OFF+i]=source[DYN2_DATA_OFF+i];
    dyn_set_checksum(destination); 
    return 0x01;
  }
}

void dyn_init_status_packet(unsigned char *packet,unsigned char id,TDynError error,unsigned char length,unsigned char *data)
{
  unsigned char i;

  packet[DYN_HEADER_OFF]=0xFF;
  packet[DYN_HEADER_OFF+1]=0xFF;
  packet[DYN_ID_OFF]=id;
  packet[DYN_LENGTH_OFF]=0x02+length;
  packet[DYN_ERROR_OFF]=error;
  for(i=0;i<length;i++)
    packet[DYN_DATA_OFF+i]=data[i];
  dyn_set_checksum(packet); 
}

/* read instruction status packet */
unsigned char dyn_get_read_status_data(unsigned char *packet,unsigned char *data)
{
  unsigned char i;

  for(i=0;i<packet[DYN_LENGTH_OFF]-0x02;i++)
    data[i]=packet[DYN_DATA_OFF+i];

  return packet[DYN_LENGTH_OFF]-0x02;
}
