#include "dynamixel_slave.h"

// communication functions
unsigned char dyn_slave_irq_receive_cb(TDynamixelSlave *dyn_slave,unsigned char byte)
{
  if(dyn_slave!=0x00000000)
  {
    if(dyn_slave->comm_dev->time!=0x00000000)
      time_set_timeout(dyn_slave->comm_dev->time,dyn_slave->rx_timeout_ms*1000);
    switch(dyn_slave->received_bytes)
    {
      case 0: if(byte==0xFF)
              {
                dyn_slave->rx_buffer[dyn_slave->received_bytes]=byte;
	        dyn_slave->received_bytes++;
              }
              break;
      case 1: if(byte==0xFF)
              {
                dyn_slave->rx_buffer[dyn_slave->received_bytes]=byte;
	        dyn_slave->received_bytes++;
              }
              else 
                dyn_slave->received_bytes--;
	      break;
      case 2: if(byte==0xFD)// version 2 header
              {
                if(dyn_slave->version==DYN_VER2)// the module is configured for version 2
                {
                  dyn_slave->rx_buffer[dyn_slave->received_bytes]=byte;
                  dyn_slave->received_bytes++;
                }
                else
                  dyn_slave->received_bytes=0;// ignore packet and restart synchronization
              } 
              else if(byte!=0xFF)
              {
                dyn_slave->rx_buffer[dyn_slave->received_bytes]=byte;
	        dyn_slave->received_bytes++;
              }
	      break;
      case 3: dyn_slave->rx_buffer[dyn_slave->received_bytes]=byte;
              if(dyn_slave->version==DYN_VER1)
              {
                dyn_slave->op_length=byte;
                dyn_slave->received_bytes=0;
                /* finish reception by IRQ */
                comm_cancel_irq_receive(dyn_slave->comm_dev);
                /* enable dma RX */
                comm_receive_dma(dyn_slave->comm_dev,&dyn_slave->rx_buffer[4],dyn_slave->op_length);
              }
              else
                dyn_slave->received_bytes++;
	      break;
      case 4: dyn_slave->rx_buffer[dyn_slave->received_bytes]=byte;
              dyn_slave->received_bytes++;
              break;
      case 5: dyn_slave->rx_buffer[dyn_slave->received_bytes]=byte;
              dyn_slave->received_bytes++;
              dyn_slave->op_length=byte;
              break;
      case 6: dyn_slave->rx_buffer[dyn_slave->received_bytes]=byte;
              dyn_slave->received_bytes++;
              dyn_slave->op_length+=(byte<<8);
              dyn_slave->received_bytes=0;
              /* finish reception by IRQ */
              comm_cancel_irq_receive(dyn_slave->comm_dev);
              /* enable dma RX */
              comm_receive_dma(dyn_slave->comm_dev,&dyn_slave->rx_buffer[7],dyn_slave->op_length);
              break;
      default: break;
    }
  }

  return 0x00;
}

unsigned char dyn_slave_dma_send_cb(TDynamixelSlave *dyn_slave)
{
  if(dyn_slave!=0x00000000)
  {
    // enable tx interrupts
    dyn_slave->set_rx_mode();  
  }

  return 0x00;
}

unsigned char dyn_slave_dma_receive_cb(TDynamixelSlave *dyn_slave)
{
  if(dyn_slave!=0x00000000)
  {
    comm_receive_irq(dyn_slave->comm_dev,0);// reenable reception by IRQ
    if(dyn_slave->comm_dev->time!=0x00000000)
      time_cancel_timeout(dyn_slave->comm_dev->time);
    dyn_slave->packet_ready=0x01;
  }

  return 0x00;
}

void dummy_dyn_slave_set_tx_mode(void)
{

}

void dummy_dyn_slave_set_rx_mode(void)
{

}

void dummy_dyn_slave_set_baudrate(TComm *comm_dev,unsigned int baudrate)
{

}

unsigned char dummy_dyn_slave_on_relay(TDynVersion version,unsigned char *inst_pkt,unsigned char *status_pkt)
{
  return DYN_NO_DEVICE;
}

void dyn_slave_send_status_packet(TDynamixelSlave *slave,unsigned char address,unsigned char error,unsigned short int length, unsigned char *data,unsigned char return_delay)
{
  if(slave!=0x00000000)
  {
    // wait until the previous transmission has ended (if any)
    while(comm_is_send_done(slave->comm_dev)==COMM_BUSY);
    if(return_delay>0)
      if(slave->comm_dev->time!=0x00000000)
        time_delay_us(slave->comm_dev->time,return_delay<<1);
    if(slave->version==DYN_VER1)
    {
      // create the status packet
      dyn_init_status_packet(slave->tx_buffer,address,(TDynError)error,length,data);
      // set the tx mode, if necessary
      slave->set_tx_mode();
      // start transmission by DMA
      comm_send_dma(slave->comm_dev,slave->tx_buffer,dyn_get_length(slave->tx_buffer)+4);
    }
    else
    {
      // create the status packet
      dyn2_init_status_packet(slave->tx_buffer,address,(TDynError)error,length,data);
      // set the tx mode, if necessary
      slave->set_tx_mode();
      // start transmission by DMA
      comm_send_dma(slave->comm_dev,slave->tx_buffer,dyn2_get_length(slave->tx_buffer)+7);
    }
  }
}

void dyn_slave_loop(TDynamixelSlave *slave)
{
  unsigned char i,send_status,error,data[MAX_DYN_SLAVE_TX_BUFFER_LEN];
  unsigned short int length;

  if(slave!=0x00000000)
  {
    if(slave->packet_ready)// check if a new instruction packet has been received
    {
      slave->packet_ready=0x00;
      // check address
      for(i=0;i<slave->num_slave_devices;i++)
      {
        if(slave->version==DYN_VER1)
          send_status=dyn_v1_slave_loop(slave->slave_devices[i],slave->rx_buffer,&error,(unsigned char *)&length,data);
        else
          send_status=dyn_v2_slave_loop(slave->slave_devices[i],slave->rx_buffer,&error,&length,data);
        if(send_status==0x01)
          dyn_slave_send_status_packet(slave,slave->slave_devices[i]->address,error,length,data,slave->slave_devices[i]->return_delay);
      }
      if(slave->on_relay(slave->version,slave->rx_buffer,slave->tx_buffer)==DYN_SUCCESS)
      {   
        // set the tx mode, if necessary           
        slave->set_tx_mode();
        // start transmission by DMA
        comm_send_dma(slave->comm_dev,slave->tx_buffer,dyn_get_length(slave->tx_buffer)+4);         
      } 
    }
    else
    {
      if(slave->comm_dev->time!=0x00000000)
      {
        if(time_is_timeout(slave->comm_dev->time))
        { 
          /* cancel any IRQ or DMA reception */
          comm_cancel_irq_receive(slave->comm_dev);
          comm_cancel_dma_receive(slave->comm_dev);
          slave->received_bytes=0;
          /* enable reception by IRQ */
          comm_receive_irq(slave->comm_dev,0);
        }
      }
    }
  }
}

/* public functions */
unsigned char dyn_slave_init(TDynamixelSlave *slave,TComm *dev,TScheduler *scheduler,sched_channel_t ch,TDynVersion version)
{
  unsigned char i;

  /* assign communication functions */
  dev->irq_receive_cb=(unsigned char (*)(void *,unsigned char))dyn_slave_irq_receive_cb;
  dev->dma_send_cb=(unsigned char (*)(void *))dyn_slave_dma_send_cb;
  dev->dma_receive_cb=(unsigned char (*)(void *))dyn_slave_dma_receive_cb;
  slave->comm_dev=dev;
  slave->version=version;
  dev->data=slave;
  /* initialize the internal callbacks */
  slave->set_tx_mode=dummy_dyn_slave_set_tx_mode;
  slave->set_rx_mode=dummy_dyn_slave_set_rx_mode;
  slave->set_baudrate=dummy_dyn_slave_set_baudrate;
  slave->on_relay=dummy_dyn_slave_on_relay;
  /* initialize internal variables */
  slave->packet_ready=0x00;
  slave->rx_timeout_ms=50;
  slave->received_bytes=0x00;
  /* initialize slave devices */
  slave->num_slave_devices=0;
  for(i=0;i<MAX_NUM_SLAVE_DEVICES;i++)
    slave->slave_devices[i]=0x00000000;

  /* assigna a scheduler channel */
  scheduler_set_channel(scheduler,ch,(void (*)(void *))dyn_slave_loop,1,slave);
  slave->scheduler=scheduler;
  slave->sch_channel=ch;

  slave->set_rx_mode();
  /* enable reception by IRQ */
  comm_receive_irq(slave->comm_dev,0);

  return 0x01;
}

unsigned char dyn_slave_add_device(TDynamixelSlave *slave, TDynamixelSlaveDevice *device)
{
  if(slave!=0x00000000)
  {
    if(slave->num_slave_devices<=(MAX_NUM_SLAVE_DEVICES-1))
    {
      slave->slave_devices[slave->num_slave_devices]=device;
      slave->num_slave_devices++;
      device->slave=slave;
      return 0x01;
    }
    else
      return 0x00;
  }
  else
    return 0x00;
}
