#include "dynamixel_slave_device.h"
#include "dynamixel_slave.h"

/* private functions */
void dyn_slave_device_write_cmd(TDynamixelSlaveDevice *device,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned short int offset;

  offset=address-device->eeprom_base_address1;
  if(ram_in_window(device->eeprom_base_address1+DEVICE_MODEL,2,address,length))
  {
    if(ram_in_range(device->eeprom_base_address1+DEVICE_MODEL,address,length))
      device->memory.data[device->eeprom_base_address1+DEVICE_MODEL]=data[DEVICE_MODEL-offset];
    if(ram_in_range(device->eeprom_base_address1+DEVICE_MODEL+1,address,length))
      device->memory.data[device->eeprom_base_address1+DEVICE_MODEL+1]=data[DEVICE_MODEL+1-offset];
  }
  if(ram_in_range(device->eeprom_base_address1+DEVICE_ID,address,length))
  {
    dyn_slave_device_set_address(device,data[DEVICE_ID-offset]);
    device->memory.data[device->eeprom_base_address1+DEVICE_ID]=data[DEVICE_ID-offset];
  }
  if(ram_in_range(device->eeprom_base_address1+BAUDRATE,address,length))
  {
    dyn_slave_set_baudrate(device->slave,2000000/(data[BAUDRATE-offset]+1));
    device->memory.data[device->eeprom_base_address1+BAUDRATE]=data[BAUDRATE-offset];
  }
  if(ram_in_range(device->eeprom_base_address1+RETURN_DELAY,address,length))
  {
    dyn_slave_device_set_return_delay(device,data[RETURN_DELAY-offset]);
    device->memory.data[device->eeprom_base_address1+RETURN_DELAY]=data[RETURN_DELAY-offset];
  }
  offset=address-device->eeprom_base_address2;
  if(ram_in_range(device->eeprom_base_address2+RETURN_LEVEL,address,length))
  {
    dyn_slave_device_set_return_level(device,(return_level_t)data[RETURN_LEVEL-offset]);
    device->memory.data[device->eeprom_base_address2+RETURN_LEVEL]=data[RETURN_LEVEL-offset];
  }
}

void dyn_slave_device_read_cmd(TDynamixelSlaveDevice *device,unsigned short int address,unsigned short int length,unsigned char *data)
{
  ram_read_table(&device->memory,address,length,data);
}

void dummy_dyn_slave_device_on_ping(void)
{

}

unsigned char dummy_dyn_slave_device_on_read(unsigned short int address,unsigned short int length,unsigned char *data)
{
  return DYN_SUCCESS;
}

unsigned char dummy_dyn_slave_device_on_write(unsigned short int address,unsigned short int length,unsigned char *data)
{
  return DYN_SUCCESS;
}

unsigned char dummy_dyn_slave_device_on_reset(void)
{
  return DYN_SUCCESS;
}

unsigned char dyn_v1_slave_loop(TDynamixelSlaveDevice *device,unsigned char *rx_buffer,unsigned char *error,unsigned char *length,unsigned char *data)
{
  static unsigned char address,prev_id,id;
  unsigned char send_status=0x00;

  if(device!=0x000000000)
  {
    id=dyn_get_id(rx_buffer);
    if(id==device->address || id==DYN_BROADCAST_ID)// the packet is addressed to this device or it is a broadcast
    {
      // check the packet checksum
      if(dyn_check_checksum(rx_buffer)==0xFF)// the incomming packet is okay
      {
        // process the packet
        switch(dyn_get_instruction(rx_buffer))
        {
          case DYN_PING: device->on_ping();
                         if(id!=DYN_BROADCAST_ID)
                         {
                           send_status=0x01;
                           *error=DYN_NO_ERROR;
                           *length=0;
                         }
                         break;
          case DYN_READ: (*error)=device->on_read(dyn_get_read_address(rx_buffer),dyn_get_read_length(rx_buffer),data);
                         if(device->return_level!=no_return && id!=DYN_BROADCAST_ID)
                         {
                           send_status=0x01;
                           if((*error)==DYN_NO_ERROR)
                             (*length)=dyn_get_read_length(rx_buffer);
                           else
                           {
                             (*length)=0;
                             (*error)=DYN_INST_ERROR;
                           }
                         }
                         break;
          case DYN_WRITE: (*length)=dyn_get_write_data(rx_buffer,data);
                          (*error)=device->on_write(dyn_get_write_address(rx_buffer),*length,data);
                          if(device->return_level==return_all && id!=DYN_BROADCAST_ID)
                          {
                            send_status=0x01;
                            (*length)=0;
                            if((*error)!=DYN_NO_ERROR)
                              (*error)=DYN_INST_ERROR;
                          }
                          break;
          case DYN_REG_WRITE: device->reg_length=dyn_get_reg_write_data(rx_buffer,device->reg_buffer);
                              device->reg_address=dyn_get_reg_write_address(rx_buffer);
                              if(device->return_level==return_all && id!=DYN_BROADCAST_ID)
                              {
                                send_status=0x01;
                                (*error)=DYN_NO_ERROR;
                                (*length)=0;
                              }
                              break;
          case DYN_ACTION: if(device->reg_address!=0xFFFF)
                           {
                             (*error)=device->on_write(device->reg_address,device->reg_length,device->reg_buffer);
                             device->reg_address=0xFFFF;
                           }
                           else
                             if(device->return_level==return_all && id!=DYN_BROADCAST_ID)
                             {
                               send_status=0x01;
                               (*error)=DYN_NO_ERROR;
                               (*length)=0;
                             }
                           break;
          case DYN_RESET:
                          break;
          case DYN_SYNC_READ: send_status=0x01;
                              (*error)=DYN_INST_ERROR;
                              (*length)=0;
                              break;
          case DYN_SYNC_WRITE: if(dyn_sync_write_id_present(rx_buffer,device->address,&address,length,data))// the device is addressed
                                 (*error)=device->on_write(address,*length,data);
                               break;
          case DYN_BULK_READ: prev_id=dyn_bulk_read_id_present(rx_buffer,device->address,&address,length);
                              if(prev_id!=0xFF)
                              {
                                if(prev_id==0x00)// first device to answer
                                {
                                  send_status=0x01;
                                  (*error)=device->on_read(address,*length,data);
                                  if((*error)!=DYN_NO_ERROR)
                                  {
                                    (*error)=DYN_INST_ERROR;
                                    (*length)=0;
                                  }
                                }
                                else// wait for the previous device in the sequence to send its data
                                {
                                  device->sync_bulk_address=address;
                                  device->sync_bulk_length=*length;
                                  device->sync_bulk_prev_id=prev_id;
                                  device->bulk_read_pending=0x01;       
                                }
                              }
                              break;
          case DYN_BULK_WRITE: send_status=0x01;
                               (*error)=DYN_INST_ERROR;
                               (*length)=0;
                               break;
          default:
                   break;
        }
      }
      else
      {
        // send a checksum error answer
        if(dyn_get_id(rx_buffer)!=DYN_BROADCAST_ID)
        {
          send_status=0x01;
          (*error)=DYN_CHECKSUM_ERROR;
          (*length)=0;
        }
      }
    }
    else
    {
      if(device->bulk_read_pending)
      {
        if(id==device->sync_bulk_prev_id)
        {
          device->bulk_read_pending=0x00;
          send_status=0x01;
          (*error)=device->on_read(device->sync_bulk_address,device->sync_bulk_length,data);
          if((*error)==DYN_NO_ERROR)
            (*length)=device->sync_bulk_length;
          else
          {
            (*error)=DYN_INST_ERROR;
            (*length)=0;
          }
        }
      }
    }
  }
  
  return send_status;
}

unsigned char dyn_v2_slave_loop(TDynamixelSlaveDevice *device,unsigned char *rx_buffer,unsigned char *error,unsigned short int *length,unsigned char *data)
{
  static unsigned char prev_id,id;
  static unsigned short int address;
  unsigned char send_status=0x00;

  if(device!=0x00000000)
  {
    id=dyn2_get_id(rx_buffer);
    if(id==device->address || id==DYN_BROADCAST_ID)// the packet is addressed to this device or it is a broadcast
    {
      // check the packet checksum
      if(dyn2_check_checksum(rx_buffer)==0x01)// the incomming packet is okay
      {
        // process the packet
        switch(dyn2_get_instruction(rx_buffer))
        {
          case DYN_PING: device->on_ping();
                         if(id!=DYN_BROADCAST_ID)
                         {
                           send_status=0x01;
                           (*error)=DYN_NO_ERROR;
                           (*length)=0;
                         }
                         break;
          case DYN_READ: (*error)=device->on_read(dyn2_get_read_address(rx_buffer),dyn2_get_read_length(rx_buffer),data);
                         if(device->return_level!=no_return && id!=DYN_BROADCAST_ID)
                         {
                           send_status=0x01;
                           if((*error)==DYN_NO_ERROR)
                             (*length)=dyn2_get_read_length(rx_buffer);
                           else
                           {
                             (*length)=0;
                             (*error)=DYN_INST_ERROR;
                           }
                         }
                         break;
          case DYN_WRITE: (*length)=dyn2_get_write_data(rx_buffer,data);
                          (*error)=device->on_write(dyn2_get_write_address(rx_buffer),*length,data);
                          if(device->return_level==return_all && id!=DYN_BROADCAST_ID)
                          {
                            send_status=0x01;
                            (*length)=0;
                            if((*error)!=DYN_NO_ERROR)
                              (*error)=DYN_INST_ERROR;
                          }
                          break;
          case DYN_REG_WRITE: device->reg_length=dyn2_get_reg_write_data(rx_buffer,device->reg_buffer);
                              device->reg_address=dyn2_get_reg_write_address(rx_buffer);
                              if(device->return_level==return_all && id!=DYN_BROADCAST_ID)
                              {
                                send_status=0x01;
                                (*error)=DYN_NO_ERROR;
                                (*length)=0;
                              }
                              break;
          case DYN_ACTION: if(device->reg_address!=0xFFFF)
                           {
                             (*error)=device->on_write(device->reg_address,device->reg_length,device->reg_buffer);
                             device->reg_address=0xFFFF;
                           }
                           else
                             if(device->return_level==return_all && id!=DYN_BROADCAST_ID)
                             {
                               send_status=0x01;
                               (*error)=DYN_NO_ERROR;
                               (*length)=0;
                             }
                           break;
          case DYN_RESET:
                          break;
          case DYN_SYNC_READ: prev_id=dyn2_sync_read_id_present(rx_buffer,device->address,&address,length);
                              if(prev_id!=0xFF)
                              {
                                if(prev_id==0x00)// first device to answer
                                {
                                  (*error)=device->on_read(address,*length,data);
                                  send_status=0x01;
                                  if((*error)!=DYN_NO_ERROR)
                                  {
                                    (*error)=DYN_INST_ERROR;
                                    (*length)=0;
                                  }
                                }
                                else// wait for the previous device in the sequence to send its data
                                {
                                  device->sync_bulk_address=address;
                                  device->sync_bulk_length=*length;
                                  device->sync_bulk_prev_id=prev_id;
                                  device->sync_read_pending=0x01;
                                }
                              }
                              break;
          case DYN_SYNC_WRITE: if(dyn2_sync_write_id_present(rx_buffer,device->address,&address,length,data))// the device is addressed
                                 (*error)=device->on_write(address,*length,data);
                               break;
          case DYN_BULK_READ: prev_id=dyn2_bulk_read_id_present(rx_buffer,device->address,&address,length);
                              if(prev_id!=0xFF)
                              {
                                if(prev_id==0x00)// first device to answer
                                {
                                  send_status=0x01;
                                  (*error)=device->on_read(address,*length,data);
                                  if((*error)!=DYN_NO_ERROR)
                                  {
                                    (*error)=DYN_INST_ERROR;
                                    (*length)=0;   
                                  }
                                }
                                else// wait for the previous device in the sequence to send its data
                                {
                                  device->sync_bulk_address=address;
                                  device->sync_bulk_length=*length;
                                  device->sync_bulk_prev_id=prev_id;
                                  device->bulk_read_pending=0x01;       
                                }
                              }
                              break;
          case DYN_BULK_WRITE: if(dyn2_bulk_write_id_present(rx_buffer,device->address,&address,length,data))
                                 (*error)=device->on_write(address,*length,data);
                               break;
          default:
                   break;
        }
      }
      else
      {
        // send a checksum error answer
        if(dyn_get_id(rx_buffer)!=DYN_BROADCAST_ID)
        {
          send_status=0x01;
          (*error)=DYN_CHECKSUM_ERROR;
          (*length)=0;
        }
      }
    }
    else
    {
      if(device->bulk_read_pending)
      {
        if(id==device->sync_bulk_prev_id)
        {
          device->bulk_read_pending=0x00;
          (*error)=device->on_read(device->sync_bulk_address,device->sync_bulk_length,data);
          send_status=0x01;
          if((*error)==DYN_NO_ERROR)
            (*length)=device->sync_bulk_length;
          else
          {
            (*length)=0;
            (*error)=DYN_INST_ERROR;
          }
        }
      }
      else if(device->sync_read_pending)
      {
        if(id==device->sync_bulk_prev_id)
        {
          device->sync_read_pending=0x00;
          (*error)=device->on_read(device->sync_bulk_address,device->sync_bulk_length,data);
          send_status=0x01;
          if((*error)==DYN_NO_ERROR)
            (*length)=device->sync_bulk_length;
          else
          {
            (*length)=0;
            (*error)=DYN_INST_ERROR;
          }
        }
      }
    }
  }

  return send_status;
}

/* public functions */
unsigned char dyn_slave_device_init(TDynamixelSlaveDevice *device,unsigned char address,unsigned char eeprom_base_address1,unsigned char eeprom_base_address2)
{
  /* initialize the internal callbacks */
  device->on_ping=dummy_dyn_slave_device_on_ping;
  device->on_read=dummy_dyn_slave_device_on_read;
  device->on_write=dummy_dyn_slave_device_on_write;
  device->on_reset=dummy_dyn_slave_device_on_reset;
  /* initialize internal variables */
  device->address=address;
  device->return_delay=0x00;
  device->return_level=return_all;
  device->reg_address=0xFFFF;
  device->reg_length=0x0000;
  device->sync_bulk_address=0x0000;
  device->sync_bulk_length=0x0000;
  device->sync_bulk_prev_id=0x00;
  device->sync_read_pending=0x00;
  device->bulk_read_pending=0x00;

  /* initialize memory module */
  mem_init(&device->memory);
  mem_module_init(&device->mem_module);
  device->mem_module.data=device;
  device->mem_module.write_cmd=(void(*)(void *,unsigned short int, unsigned short int,unsigned char *))dyn_slave_device_write_cmd;
  device->mem_module.read_cmd=(void(*)(void *,unsigned short int, unsigned short int,unsigned char *))dyn_slave_device_read_cmd;
  if(!mem_module_add_eeprom_segment(&device->mem_module,eeprom_base_address1,EEPROM_DYN_SLAVE_LENGTH1))
    return 0x00;
  if(!mem_module_add_eeprom_segment(&device->mem_module,eeprom_base_address2,EEPROM_DYN_SLAVE_LENGTH2))
    return 0x00;
  if(!mem_add_module(&device->memory,&device->mem_module))
    return 0x00;
  device->eeprom_base_address1=eeprom_base_address1;
  device->eeprom_base_address2=eeprom_base_address2;

  return 0x01;
}
