#ifndef _DYN_MODULE_H
#define _DYN_MODULE_H

#ifdef __cplusplus
extern "C" {
#endif

#ifndef MODULE_MAX_NUM_MODELS
  #error "Please, specify the maximum number of dynamixel master modules with the MODULE_MAX_NUM_MODELS macro"
#endif

#include "dyn_manager.h"
#include "memory.h"

typedef struct TDynModule{
  TDynManager *manager;
  unsigned short int models[MODULE_MAX_NUM_MODELS];
  unsigned char num_models;
  unsigned char assigned_ids[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned char num_assigned_ids;
  unsigned char period_count;
  unsigned char enabled;
  void (*add_device)(void *module_data,unsigned char id,unsigned short int model);
  void (*set_period)(void *module_data,unsigned short int period_ms);
  void (*setup)(void *module_data);
  void (*pre_process)(void *module_data);
  void (*post_process)(void *module_data);
  void *data;
  TMemory *memory;
  TMemModule mem_module;
  unsigned short int ram_base_address;
  unsigned short int eeprom_base_address;
}TDynModule;

//public functions
unsigned char dyn_module_init(TDynModule *module,TMemory *memory,unsigned short int ram_base_address, unsigned short int eeprom_base_address);
static inline void dyn_module_enable(TDynModule *module)
{
  module->enabled=0x01;
}
static inline void dyn_module_disable(TDynModule *module)
{
  module->enabled=0x00;
}
static inline unsigned char dyn_module_is_enabled(TDynModule *module)
{
  return module->enabled;
}
void dyn_module_add_model(TDynModule *module,unsigned short int model);
static inline unsigned char dyn_module_get_num_models(TDynModule *module)
{
  return module->num_models;
}
void dyn_module_add_device(TDynModule *module,unsigned char id, unsigned short int model);
static inline unsigned char dyn_module_get_num_assigned_devices(TDynModule *module)
{
  return module->num_assigned_ids;
}
unsigned char dyn_module_device_is_assigned(TDynModule *module,unsigned char id);
void dyn_module_set_period(TDynModule *module,unsigned char period);
static inline unsigned char dyn_module_get_period(TDynModule *module)
{
  return module->period_count;
}

#ifdef __cplusplus
  }
#endif

#endif
