#ifndef _MOTION_MANAGER_H
#define _MOTION_MANAGER_H

#ifdef __cplusplus
extern "C" {
#endif

#include "dyn_module.h"
#include "dyn_servos.h"
#include "motion_manager_registers.h"

#ifndef MM_MAX_NUM_MOTION_MODULES
  #error "Please, specify the maximum number of motion modules with the MM_MAX_NUM_MOTION_MODULES macro"
#endif

typedef enum {MM_NONE          = 7,
              MM_ACTION        = 0,
              MM_WALKING       = 1,
              MM_JOINTS        = 2,
              MM_HEAD          = 3} TModules;

typedef struct{
  unsigned char cw_compliance;
  unsigned char ccw_compliance;
  unsigned short int target_value;
  long long int target_angle;
  unsigned int current_value;
  long long int current_angle;
  TModules module;
  unsigned char enabled;
  char offset;
}TServoValues;

struct TMotionModule;

typedef struct {
  TDynModule dyn_module;
  struct TMotionModule *modules[MM_MAX_NUM_MOTION_MODULES];
  unsigned char num_modules;
  TServoConfig *servo_configs[DYN_MANAGER_MAX_NUM_DEVICES];
  TServoValues servo_values[DYN_MANAGER_MAX_NUM_DEVICES];
  OP_HANDLE *enable_op;
  OP_HANDLE *motion_op[2];
  OP_HANDLE *feedback_op[2];
  unsigned char balance_enabled;
  void (*balance)(short int offsets[DYN_MANAGER_MAX_NUM_DEVICES]);
  TMemory *memory;
  TMemModule mem_module;
  unsigned short int ram_base_address;
  unsigned short int eeprom_base_address;
  unsigned int present_servos;
}TMotionManager;  

unsigned char mmanager_init(TMotionManager *mmanager,TMemory *memory,unsigned short int ram_base_address,unsigned short int eeprom_base_address);
static inline TDynModule *mmanager_get_dyn_module(TMotionManager *mmanager)
{
  return &mmanager->dyn_module;
}
static inline void mmanager_enable_balance(TMotionManager *mmanager)
{
  if(mmanager->balance!=0x00000000)
    mmanager->balance_enabled=0x01;
}
static inline void mmanager_disable_balance(TMotionManager *mmanager)
{
  mmanager->balance_enabled=0x00;
}
static inline unsigned char mmanager_is_balance_enabled(TMotionManager *mmanager)
{
  return mmanager->balance_enabled;
}
void mmanager_set_module(TMotionManager *mmanager,unsigned char servo_id,TModules module);
TModules mmanager_get_module(TMotionManager *mmanager,unsigned char servo_id);
void mmanager_enable_servo(TMotionManager *mmanager,unsigned char servo_id);
void mmanager_disable_servo(TMotionManager *mmanager,unsigned char servo_id);
unsigned char mmanager_is_servo_enabled(TMotionManager *mmanager,unsigned char servo_id);
static inline void mmanager_set_offset(TMotionManager *mmanager,unsigned char servo_id,char offset)
{
  if(servo_id<DYN_MANAGER_MAX_NUM_DEVICES)
    mmanager->servo_values[servo_id].offset=offset;
}
void mmanager_add_module(TMotionManager *mmanager,struct TMotionModule *module);

#ifdef __cplusplus
  }
#endif

#endif
