#ifndef _MOTION_PAGES_H
#define _MOTION_PAGES_H

#ifdef __cplusplus
extern "C" {
#endif

#define MAX_PAGES                     256 
#define PAGE_MAX_NUM_SERVOS           31
#define POSE_NUMBER_OF_POSES_PER_PAGE 7

typedef struct // Header Structure (total 64unsigned char)
{
  unsigned char name[14];         // Name             0~13
  unsigned char reserved1;        // Reserved1        14
  unsigned char repeat;           // Repeat count     15
  unsigned char schedule;         // schedule         16
  unsigned char reserved2[3];     // reserved2        17~19
  unsigned char stepnum;          // Number of step   20
  unsigned char reserved3;        // reserved3        21
  unsigned char speed;            // Speed            22
  unsigned char reserved4;        // reserved4        23
  unsigned char accel;            // Acceleration time 24
  unsigned char next;             // Link to next     25
  unsigned char exit;             // Link to exit     26
  unsigned char reserved5[4];     // reserved5        27~30
  unsigned char checksum;         // checksum         31
  unsigned char slope[PAGE_MAX_NUM_SERVOS];        // CW/CCW compliance slope  32~62
  unsigned char reserved6;        // reserved6        63
}TPageHeader;

typedef struct // Step Structure (total 64unsigned char)
{
  short int position[PAGE_MAX_NUM_SERVOS];    // Joint position   0~61
  unsigned char pause;            // Pause time       62
  unsigned char time;             // Time             63
}TStep;

typedef struct // Page Structure (total 512unsigned char)
{
  TPageHeader header;       // Page header  0~64
  TStep steps[POSE_NUMBER_OF_POSES_PER_PAGE];           // Page step    65~511
}TPage;

extern TPage motion_pages[MAX_PAGES];

// public functions
void pages_get_page(unsigned char page_id,TPage *page);
unsigned char pages_check_checksum(TPage *page);
void pages_clear_page(TPage *page);
void pages_copy_page(TPage *src,TPage *dst);
unsigned char pages_get_slope(TPage *page,unsigned char servo_id);

#ifdef __cplusplus
}
#endif

#endif
