#include "dyn_manager.h"
#include "dyn_module.h"
#include "ram.h"

extern unsigned char dyn_master_start_read_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data);
extern unsigned char dyn_master_is_read_table_done(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data);
extern unsigned char dyn_master_start_write_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data);
extern unsigned char dyn_master_is_write_table_done(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data);
extern unsigned char dyn_master_start_sync_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[]);
extern unsigned char dyn_master_is_sync_write_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[]);
extern unsigned char dyn_master_start_sync_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[]);
extern unsigned char dyn_master_is_sync_read_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[]);
extern unsigned char dyn_master_start_bulk_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[]);
extern unsigned char dyn_master_is_bulk_write_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[]);
extern unsigned char dyn_master_start_bulk_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[]);
extern unsigned char dyn_master_is_bulk_read_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[]);

typedef enum {start_single_ops,wait_single_ops,start_sync_ops,wait_sync_ops,start_bulk_ops,wait_bulk_ops,ops_done} dyn_manager_states_t;

// private functions
void dyn_manager_start_scan(TDynManager *manager)
{
  if(manager->running==0x00)
    dyn_manager_start(manager);
  manager->do_scan=0x01;
  manager->memory->data[manager->ram_base_address+DYN_MANAGER_CONTROL_OFFSET]|=DYN_MANAGER_SCANNING;
}

void dyn_manager_write_cmd(TDynManager *module,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned short int eeprom_offset,ram_offset;

  eeprom_offset=address-module->eeprom_base_address;
  if(ram_in_window(module->eeprom_base_address+DYN_MANAGER_PERIOD_OFFSET,1,address,length))
    dyn_manager_set_period(module,data[DYN_MANAGER_PERIOD_OFFSET-eeprom_offset]);

  ram_offset=address-module->ram_base_address;
  if(ram_in_range(module->ram_base_address+DYN_MANAGER_CONTROL_OFFSET,address,length))
  {
    if(data[DYN_MANAGER_CONTROL_OFFSET-ram_offset]&DYN_MANAGER_START)
      dyn_manager_start(module);
    if(data[DYN_MANAGER_CONTROL_OFFSET-ram_offset]&DYN_MANAGER_STOP)
      dyn_manager_stop(module);
    if(data[DYN_MANAGER_CONTROL_OFFSET-ram_offset]&DYN_MANAGER_START_SCAN)
      dyn_manager_start_scan(module);
  }
}

void dyn_manager_read_cmd(TDynManager *module,unsigned short int address,unsigned short int length,unsigned char *data)
{
  ram_read_table(module->memory,address,length,data);
}

unsigned char dyn_manager_check_id(TDynManager *manager,unsigned char id)
{
  if(id>=DYN_MANAGER_MAX_NUM_DEVICES)
    return 0xFF;
  if(manager->devices[id].model==0xFFFF)
    return 0xFF;
  else
    return manager->devices[id].master_index;
}

TDynManagerOpCommon *dyn_manager_get_op_common(TDynManager *manager,unsigned char master_index,OP_HANDLE *op)
{
  if(op->op_index[master_index]!=0xFF)
  {
    switch(op->op_type&DYN_MANAGER_OP_MASK)
    {
      case DYN_MANAGER_SINGLE: return &manager->operations[master_index].single_op[op->op_index[master_index]].common;
      case DYN_MANAGER_SYNC: return &manager->operations[master_index].sync_op[op->op_index[master_index]].common;
      case DYN_MANAGER_BULK: return &manager->operations[master_index].bulk_op[op->op_index[master_index]].common;
      default: return 0x00000000;
    } 
  }
  else return 0x00000000;
}

OP_HANDLE *dyn_manager_get_free_op_handle(TDynManager *manager)
{
  unsigned char i;

  for(i=0;i<DYN_MANAGER_MAX_NUM_OP;i++)
  {
    if(manager->op_handles[i].op_type==no_op)
      return &manager->op_handles[i];
  }
  return 0x00000000;
}

unsigned char dyn_manager_id_present(unsigned char num,unsigned char *ids,unsigned char new_id)
{
  unsigned char i;

  for(i=0;i<num;i++)
  {
    if(ids[i]==new_id)
      return i;
  }
  return 0x00;
}

void dyn_manager_delete_op_single_master(TDynManager *manager,unsigned char master_index,OP_HANDLE *op)
{
  unsigned char j,k,del;

  if(op->op_index[master_index]!=0xFF)
  {
    switch(op->op_type&DYN_MANAGER_OP_MASK)
    {
      case DYN_MANAGER_SINGLE: if(op->op_index[master_index]<(DYN_MANAGER_MAX_NUM_SINGLE_OP-1))
                               {
                                 for(j=op->op_index[master_index];j<manager->operations[master_index].num_single_op;j++)
                                 {
                                   manager->operations[master_index].single_op[j].common.operation=manager->operations[master_index].single_op[j+1].common.operation;
                                   manager->operations[master_index].single_op[j].common.repetitions=manager->operations[master_index].single_op[j+1].common.repetitions;
                                   manager->operations[master_index].single_op[j].common.period_count=manager->operations[master_index].single_op[j+1].common.period_count;
                                   manager->operations[master_index].single_op[j].common.op_handle=manager->operations[master_index].single_op[j+1].common.op_handle;
                                   manager->operations[master_index].single_op[j].id=manager->operations[master_index].single_op[j+1].id;
                                   manager->operations[master_index].single_op[j].address=manager->operations[master_index].single_op[j+1].address;
                                   manager->operations[master_index].single_op[j].length=manager->operations[master_index].single_op[j+1].length;
                                   manager->operations[master_index].single_op[j].data=manager->operations[master_index].single_op[j+1].data;
                                 }
                               }
                               else 
                               {
                                 manager->operations[master_index].single_op[op->op_index[master_index]].common.operation=no_op;
                                 manager->operations[master_index].single_op[op->op_index[master_index]].common.repetitions=0x01;
                                 manager->operations[master_index].single_op[op->op_index[master_index]].common.period_count=0x01;
                                 manager->operations[master_index].single_op[op->op_index[master_index]].common.op_handle=0x00000000;
                                 manager->operations[master_index].single_op[op->op_index[master_index]].id=0xFF;
                                 manager->operations[master_index].single_op[op->op_index[master_index]].address=0x0000;
                                 manager->operations[master_index].single_op[op->op_index[master_index]].length=0x0000;
                                 manager->operations[master_index].single_op[op->op_index[master_index]].data=0x00000000;
                               }
                               manager->operations[master_index].num_single_op--;
                               break;
      case DYN_MANAGER_SYNC: if(op->op_index[master_index]<(DYN_MANAGER_MAX_NUM_SYNC_OP-1))
                             {
                               for(j=op->op_index[master_index];j<manager->operations[master_index].num_sync_op;j++)
                               {
                                 manager->operations[master_index].sync_op[j].common.operation=manager->operations[master_index].sync_op[j+1].common.operation;
                                 manager->operations[master_index].sync_op[j].common.repetitions=manager->operations[master_index].sync_op[j+1].common.repetitions;
                                 manager->operations[master_index].sync_op[j].common.period_count=manager->operations[master_index].sync_op[j+1].common.period_count;
                                 manager->operations[master_index].sync_op[j].common.op_handle=manager->operations[master_index].sync_op[j+1].common.op_handle;
                                 manager->operations[master_index].sync_op[j].address=manager->operations[master_index].sync_op[j+1].address;
                                 manager->operations[master_index].sync_op[j].length=manager->operations[master_index].sync_op[j+1].length;
                                 manager->operations[master_index].sync_op[j].num_devices=manager->operations[master_index].sync_op[j+1].num_devices;
                                 for(k=0;k<manager->operations[master_index].sync_op[j].num_devices;k++)
                                 {
                                   manager->operations[master_index].sync_op[j].ids[k]=manager->operations[master_index].sync_op[j+1].ids[k];
                                   manager->operations[master_index].sync_op[j].data[k]=manager->operations[master_index].sync_op[j+1].data[k];
                                 }
                               }
                             }
                             else 
                             {
                               manager->operations[master_index].sync_op[op->op_index[master_index]].common.operation=no_op;
                               manager->operations[master_index].sync_op[op->op_index[master_index]].common.repetitions=0x01;
                               manager->operations[master_index].sync_op[op->op_index[master_index]].common.period_count=0x01;
                               manager->operations[master_index].sync_op[op->op_index[master_index]].common.op_handle=0x00000000;
                               manager->operations[master_index].sync_op[op->op_index[master_index]].address=0x0000;
                               manager->operations[master_index].sync_op[op->op_index[master_index]].length=0x0000;
                               manager->operations[master_index].sync_op[op->op_index[master_index]].num_devices=0x00;
                               for(k=0;k<manager->operations[master_index].sync_op[op->op_index[master_index]].num_devices;k++)
                               {
                                 manager->operations[master_index].sync_op[op->op_index[master_index]].ids[k]=0xFF;
                                 manager->operations[master_index].sync_op[op->op_index[master_index]].data[k]=0x00000000;
                               }
                             }
                             manager->operations[master_index].num_sync_op--;
                             break;
      case DYN_MANAGER_BULK: if(op->op_index[master_index]<(DYN_MANAGER_MAX_NUM_BULK_OP-1))
                             {
                               for(j=op->op_index[master_index];j<manager->operations[master_index].num_bulk_op;j++)
                               {
                                 manager->operations[master_index].bulk_op[j].common.operation=manager->operations[master_index].bulk_op[j+1].common.operation;
                                 manager->operations[master_index].bulk_op[j].common.repetitions=manager->operations[master_index].bulk_op[j+1].common.repetitions;
                                 manager->operations[master_index].bulk_op[j].common.period_count=manager->operations[master_index].bulk_op[j+1].common.period_count;
                                 manager->operations[master_index].bulk_op[j].common.op_handle=manager->operations[master_index].bulk_op[j+1].common.op_handle;
                                 manager->operations[master_index].bulk_op[j].num_devices=manager->operations[master_index].bulk_op[j+1].num_devices;
                                 for(k=0;k<manager->operations[master_index].bulk_op[j].num_devices;k++)
                                 {
                                   manager->operations[master_index].bulk_op[j].address[k]=manager->operations[master_index].bulk_op[j+1].address[k];
                                   manager->operations[master_index].bulk_op[j].length[k]=manager->operations[master_index].bulk_op[j+1].length[k];
                                   manager->operations[master_index].bulk_op[j].ids[k]=manager->operations[master_index].bulk_op[j+1].ids[k];
                                   manager->operations[master_index].bulk_op[j].data[k]=manager->operations[master_index].bulk_op[j+1].data[k];
                                 }
                               }
                             }
                             else
                             {
                               manager->operations[master_index].bulk_op[op->op_index[master_index]].common.operation=no_op;
                               manager->operations[master_index].bulk_op[op->op_index[master_index]].common.repetitions=0x01;
                               manager->operations[master_index].bulk_op[op->op_index[master_index]].common.period_count=0x01;
                               manager->operations[master_index].bulk_op[op->op_index[master_index]].common.op_handle=0x00000000;
                               manager->operations[master_index].bulk_op[op->op_index[master_index]].num_devices=0x00;
                               for(k=0;k<manager->operations[master_index].bulk_op[op->op_index[master_index]].num_devices;k++)
                               {
                                 manager->operations[master_index].bulk_op[op->op_index[master_index]].address[k]=0x0000;
                                 manager->operations[master_index].bulk_op[op->op_index[master_index]].length[k]=0x0000;
                                 manager->operations[master_index].bulk_op[op->op_index[master_index]].ids[k]=0xFF;
                                 manager->operations[master_index].bulk_op[op->op_index[master_index]].data[k]=0x00000000;
                               }
                             }
                             manager->operations[master_index].num_bulk_op--;
                             break;
    }
  }
  else
    op->op_index[master_index]=0xFF;
  // remove handles
  del=0x01;
  for(j=0;j<manager->num_masters;j++)
    if(op->op_index[j]!=0xFF)
      del=0x00;
  if(del==0x01)
  {
    op->op_type=no_op;
    manager->num_ops--;
  }
}

// public functions
unsigned char dyn_manager_init(TDynManager *manager,TMemory *memory,TScheduler *scheduler,sched_channel_t sch_channel,unsigned short int ram_base_address,unsigned short int eeprom_base_address)
{
  unsigned char i,j,k;

  for(i=0;i<DYN_MANAGER_MAX_NUM_MASTERS;i++)
  {
    manager->masters[i]=0x00000000;
    for(j=0;j<DYN_MANAGER_MAX_NUM_SINGLE_OP;j++)
    {
      manager->operations[i].single_op[j].common.operation=no_op;
      manager->operations[i].single_op[j].common.repetitions=DYN_MANAGER_PERIODIC;
      manager->operations[i].single_op[j].common.period_count=0x01;
      manager->operations[i].single_op[j].common.op_handle=0x00000000;
      manager->operations[i].single_op[j].id=0xFF;
      manager->operations[i].single_op[j].address=0x0000;
      manager->operations[i].single_op[j].length=0x0000;
      manager->operations[i].single_op[j].data=0x00000000;
    }
    manager->operations[i].num_single_op=0;
    for(j=0;j<DYN_MANAGER_MAX_NUM_SYNC_OP;j++)
    {
      manager->operations[i].sync_op[j].common.operation=no_op;
      manager->operations[i].sync_op[j].common.repetitions=DYN_MANAGER_PERIODIC;
      manager->operations[i].sync_op[j].common.period_count=0x01;
      manager->operations[i].sync_op[j].common.op_handle=0x00000000;
      manager->operations[i].sync_op[j].num_devices=0;
      for(k=0;k<DYN_MANAGER_MAX_NUM_DEVICES;k++)
      {
        manager->operations[i].sync_op[j].ids[k]=0xFF;
        manager->operations[i].sync_op[j].data[k]=0x00000000;
      }
      manager->operations[i].sync_op[j].address=0x0000;
      manager->operations[i].sync_op[j].length=0x0000;
    }
    manager->operations[i].num_sync_op=0;
    for(j=0;j<DYN_MANAGER_MAX_NUM_BULK_OP;j++)
    {
      manager->operations[i].bulk_op[j].common.operation=no_op;
      manager->operations[i].bulk_op[j].common.repetitions=DYN_MANAGER_PERIODIC;
      manager->operations[i].bulk_op[j].common.period_count=0x01;
      manager->operations[i].bulk_op[j].common.op_handle=0x00000000;
      manager->operations[i].bulk_op[j].num_devices=0;
      for(k=0;k<DYN_MANAGER_MAX_NUM_DEVICES;k++)
      {
        manager->operations[i].bulk_op[j].ids[k]=0xFF;
        manager->operations[i].bulk_op[j].address[k]=0x0000;
        manager->operations[i].bulk_op[j].length[k]=0x0000;
        manager->operations[i].bulk_op[j].data[k]=0x00000000;
      }
    }
    manager->operations[i].num_bulk_op=0;
  }
  manager->num_masters=0;
  for(i=0;i<DYN_MANAGER_MAX_NUM_MODULES;i++)
    manager->modules[i]=0x00000000;
  manager->num_modules=0;
  for(i=0;i<DYN_MANAGER_MAX_NUM_DEVICES;i++)
  {
    manager->devices[i].model=0xFFFF;
    manager->devices[i].master_index=0xFF;
  }
  manager->num_devices=0;
  for(i=0;i<DYN_MANAGER_MAX_NUM_OP;i++)
  {
    manager->op_handles[i].op_type=no_op;
    for(j=0;j<DYN_MANAGER_MAX_NUM_MASTERS;j++)
      manager->op_handles[i].op_index[j]=0xFF;
    manager->op_handles[i].handle_index=i;
  }
  manager->num_ops=0x00;
  manager->present_devices=0x00000000;
  /* initialize scheduler */
  scheduler_set_channel(scheduler,sch_channel,(void(*)(void *))dyn_manager_loop,100,manager);
  manager->scheduler=scheduler;
  manager->sch_channel=sch_channel;

  /* initialize memory module */
  mem_module_init(&manager->mem_module);
  manager->mem_module.data=manager;
  manager->mem_module.write_cmd=(void(*)(void *,unsigned short int, unsigned short int,unsigned char *))dyn_manager_write_cmd;
  manager->mem_module.read_cmd=(void(*)(void *,unsigned short int, unsigned short int,unsigned char *))dyn_manager_read_cmd;
  if(!mem_module_add_ram_segment(&manager->mem_module,ram_base_address,RAM_DYN_MANAGER_LENGTH))
    return 0x00;
  if(!mem_module_add_eeprom_segment(&manager->mem_module,eeprom_base_address,EEPROM_DYN_MANAGER_LENGTH))
    return 0x00;
  if(!mem_add_module(memory,&manager->mem_module))
    return 0x00;
  manager->eeprom_base_address=eeprom_base_address;
  manager->ram_base_address=ram_base_address;
  manager->memory=memory;
  manager->running=0x00;
  manager->do_scan=0x00;
  
  return 0x01;
}

void dyn_manager_set_period(TDynManager *manager,unsigned char period_ms)
{
  unsigned char i;

  scheduler_change_period(manager->scheduler,manager->sch_channel,period_ms);
  for(i=0;i<DYN_MANAGER_MAX_NUM_MODULES;i++)
    if(manager->modules[i]!=0x00000000 && manager->modules[i]->set_period!=0x00000000)
       manager->modules[i]->set_period(manager->modules[i]->data,period_ms*manager->modules[i]->period_count);
  manager->period_ms=period_ms;
  manager->memory->data[manager->eeprom_base_address+DYN_MANAGER_PERIOD_OFFSET]=period_ms;
}

void dyn_manager_start(TDynManager *manager)
{
  scheduler_enable_channel(manager->scheduler,manager->sch_channel);
  manager->running=0x01;
  manager->memory->data[manager->ram_base_address+DYN_MANAGER_CONTROL_OFFSET]|=DYN_MANAGER_RUNNING;
}

void dyn_manager_stop(TDynManager *manager)
{
  scheduler_disable_channel(manager->scheduler,manager->sch_channel);
  manager->running=0x00;
  manager->memory->data[manager->ram_base_address+DYN_MANAGER_CONTROL_OFFSET]&=(~DYN_MANAGER_RUNNING);
}

void dyn_manager_scan(TDynManager *manager)
{
  unsigned char servo_ids[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned char i,j,k,num=0;
  unsigned short int model;

  manager->present_devices=0x00000000;
  // scan all dynamixel buses
  for(i=0;i<manager->num_masters;i++)
  {
    // scan the bus
    dyn_master_scan(manager->masters[i],&num,servo_ids);
    for(j=0;j<num;j++)
    {
      // get the servo module
      dyn_master_read_word(manager->masters[i],servo_ids[j],0x0000,&model);
      // store internal info
      if(manager->devices[servo_ids[j]].master_index==0xFF && manager->devices[servo_ids[j]].model==0xFFFF)// device not yet initialized
      {
        manager->devices[servo_ids[j]].model=model;
        manager->devices[servo_ids[j]].master_index=i;
        manager->num_devices++;
        manager->present_devices|=(0x00000001<<servo_ids[j]);
        // pass the infor to each of the modules
        for(k=0;k<DYN_MANAGER_MAX_NUM_MODULES;k++)
          if(manager->modules[k]!=0x00000000)
            dyn_module_add_device(manager->modules[k],servo_ids[j],model); 
      }
      // else ignore device
    }
  }
  manager->memory->data[manager->ram_base_address+DYN_MANAGER_NUM_DEVICES_OFFSET]=manager->num_devices;
  manager->memory->data[manager->ram_base_address+DYN_MANAGER_CONTROL_OFFSET]&=(~DYN_MANAGER_SCANNING);
  manager->memory->data[manager->ram_base_address+DYN_MANAGER_PRESENT_DEVICES]=manager->present_devices&0x000000FF;
  manager->memory->data[manager->ram_base_address+DYN_MANAGER_PRESENT_DEVICES+1]=(manager->present_devices>>8)&0x000000FF;
  manager->memory->data[manager->ram_base_address+DYN_MANAGER_PRESENT_DEVICES+2]=(manager->present_devices>>16)&0x000000FF;
  manager->memory->data[manager->ram_base_address+DYN_MANAGER_PRESENT_DEVICES+3]=(manager->present_devices>>24)&0x000000FF;
  // configure the operations
  for(i=0;i<manager->num_modules;i++)
    if(manager->modules[i]->setup!=0x00000000)
      manager->modules[i]->setup(manager->modules[i]->data);
}

void dyn_manager_add_master(TDynManager *manager,TDynamixelMaster *master)
{
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
    if(manager->masters[i]==master)
      return;
  if(i<DYN_MANAGER_MAX_NUM_MASTERS)
  {
    manager->masters[i]=master;
    manager->num_masters++;
    manager->memory->data[manager->ram_base_address+DYN_MANAGER_NUM_MASTERS_OFFSET]++;
  }
}

TDynamixelMaster *dyn_manager_get_device_master(TDynManager *manager,unsigned char device_id)
{
  if(dyn_manager_check_id(manager,device_id)!=0xFF)
    return manager->masters[dyn_manager_check_id(manager,device_id)];
  else
    return 0x00000000;
}

void dyn_manager_add_module(TDynManager *manager,TDynModule *module)
{
  unsigned char i;

  for(i=0;i<manager->num_modules;i++)
    if(manager->modules[i]==module)
      return;
  if(i<DYN_MANAGER_MAX_NUM_MODULES)
  {
    manager->modules[i]=module;
    manager->num_modules++;
    module->manager=manager;
    manager->memory->data[manager->ram_base_address+DYN_MANAGER_NUM_MODULES_OFFSET]++;
    manager->modules[i]->set_period(manager->modules[i]->data,manager->period_ms*manager->modules[i]->period_count);
  }
}

void dyn_manager_delete_op(TDynManager *manager,OP_HANDLE *op)
{
  unsigned char i;

  // remove operations
  for(i=0;i<manager->num_masters;i++)
    dyn_manager_delete_op_single_master(manager,i,op);
}

void dyn_manager_set_op_period(TDynManager *manager,OP_HANDLE *op,unsigned char period_count)
{
  TDynManagerOpCommon *common;
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if((common=dyn_manager_get_op_common(manager,i,op))!=0x00000000)
    {
      if(period_count==0x00)
        common->period_count=0x01;
      else
        common->period_count=period_count;
    }
  }
}

unsigned char dyn_manager_get_op_period(TDynManager *manager,OP_HANDLE *op)
{
  TDynManagerOpCommon *common;
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if((common=dyn_manager_get_op_common(manager,i,op))!=0x00000000)
      return common->period_count;
  }
  return 0x00;
}

void dyn_manager_set_op_repetitions(TDynManager *manager,OP_HANDLE *op,unsigned char repetitions)
{
  TDynManagerOpCommon *common;
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if((common=dyn_manager_get_op_common(manager,i,op))!=0x00000000)
    {
      if(repetitions==0x00)
        common->repetitions=0x01;
      else
        common->repetitions=repetitions;
    }
  }
}

unsigned char dyn_manager_get_op_repetitions(TDynManager *manager,OP_HANDLE *op)
{
  TDynManagerOpCommon *common;
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if((common=dyn_manager_get_op_common(manager,i,op))!=0x00000000)
      return common->repetitions;
  }
  return 0x00;
}

// single operation functions
OP_HANDLE *dyn_manager_single_op_new(TDynManager *manager,unsigned char mode,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned char master_index,current_op_index;
  TDynManagerSingleOp *current_op;
  OP_HANDLE *op_handle;

  if((master_index=dyn_manager_check_id(manager,id))==0xFF)
    return 0x00000000;
  if((op_handle=dyn_manager_get_free_op_handle(manager))==0x00000000)
    return 0x00000000;
  if(mode==DYN_MANAGER_READ || mode==DYN_MANAGER_WRITE)
    op_handle->op_type=(op_type_t)(DYN_MANAGER_SINGLE|mode);
  else
    return 0x00000000;
  current_op_index=manager->operations[master_index].num_single_op;
  if(current_op_index<DYN_MANAGER_MAX_NUM_SINGLE_OP)
  {
    current_op=&manager->operations[master_index].single_op[current_op_index];
    current_op->common.operation=(op_type_t)(DYN_MANAGER_SINGLE|mode);
    current_op->common.op_handle=op_handle;
    current_op->id=id;
    current_op->address=address;
    current_op->length=length;
    current_op->data=data;
    op_handle->op_index[master_index]=current_op_index;
    manager->operations[master_index].num_single_op++;
    manager->num_ops++;
  }
  else
  {
    op_handle->op_type=no_op;
    return 0x00000000;
  }
  return op_handle;
}

void dyn_manager_single_op_change_id(TDynManager *manager,OP_HANDLE *op,unsigned char id)
{
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
      manager->operations[i].single_op[op->op_index[i]].id=id;
  }
}

void dyn_manager_single_op_change_address(TDynManager *manager,OP_HANDLE *op,unsigned short int address)
{
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
      manager->operations[i].single_op[op->op_index[i]].address=address;
  }
}

void dyn_manager_single_op_change_length(TDynManager *manager,OP_HANDLE *op,unsigned short int length)
{
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
      manager->operations[i].single_op[op->op_index[i]].length=length;
  }
}

void dyn_manager_single_op_change_data(TDynManager *manager,OP_HANDLE *op,unsigned char *data)
{
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
      manager->operations[i].single_op[op->op_index[i]].data=data;
  }
}

// sync operation functions
OP_HANDLE *dyn_manager_sync_op_new(TDynManager *manager,unsigned char mode,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  unsigned char i,master_index,current_op_index;
  TDynManagerSyncOp *current_op;
  OP_HANDLE *op_handle;

  if((op_handle=dyn_manager_get_free_op_handle(manager))==0x00000000)
    return 0x00000000;
  if(mode==DYN_MANAGER_READ || mode==DYN_MANAGER_WRITE)
    op_handle->op_type=(op_type_t)(DYN_MANAGER_SYNC|mode);
  else
    return 0x00000000;
  for(i=0;i<num;i++)
  {
    if((master_index=dyn_manager_check_id(manager,ids[i]))==0xFF)
    {
      op_handle->op_type=no_op;
      return 0x00000000;
    }
    if(op_handle->op_index[master_index]==0xFF)
    {
      current_op_index=manager->operations[master_index].num_sync_op;
      if(current_op_index<DYN_MANAGER_MAX_NUM_SYNC_OP)
      {
        current_op=&manager->operations[master_index].sync_op[current_op_index];
        if(dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[i])==0x00)
        {
          op_handle->op_index[master_index]=current_op_index;
          current_op->common.operation=(op_type_t)(DYN_MANAGER_SYNC|mode);
          current_op->common.op_handle=op_handle;
          current_op->address=address;
          current_op->length=length;
          current_op->ids[current_op->num_devices]=ids[i];
          current_op->data[current_op->num_devices]=data[i];
          current_op->num_devices++;
          manager->operations[master_index].num_sync_op++;
        }
      }
      else
      {
        op_handle->op_type=no_op;
        return 0x00000000;
      }
    }
    else
    {
      current_op_index=op_handle->op_index[master_index];
      current_op=&manager->operations[master_index].sync_op[current_op_index];
      if(dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[i])==0x00)
      {
        current_op->ids[current_op->num_devices]=ids[i];
        current_op->data[current_op->num_devices]=data[i];
        current_op->num_devices++;
      }
    }
  }
  manager->num_ops++;
  return op_handle;
}

void dyn_manager_sync_op_add_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned char * const data[])
{
  TDynManagerSyncOp *current_op;
  unsigned char i,j;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
    {
      current_op=&manager->operations[i].sync_op[op->op_index[i]];
      for(j=0;j<num;j++)
      {
        if(dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[j])==0x00)
        {
          if(current_op->num_devices<DYN_MANAGER_MAX_NUM_DEVICES)
          {
            current_op->ids[current_op->num_devices]=ids[i];
            current_op->data[current_op->num_devices]=data[i];
            current_op->num_devices++;
          }
        }
      }
    }
  }
}

void dyn_manager_sync_op_remove_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids)
{
  TDynManagerSyncOp *current_op;
  unsigned char i,j,k,device_index;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
    {
      current_op=&manager->operations[i].sync_op[op->op_index[i]];
      for(j=0;j<num;j++)
      {
        if((device_index=dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[j]))!=0x00)
        {
          if(device_index<(DYN_MANAGER_MAX_NUM_DEVICES-1))
          {
            for(k=device_index;k<current_op->num_devices;k++)
            {
              current_op->ids[k]=current_op->ids[k+1];
              current_op->data[k]=current_op->data[k+1];
            }
          }
          else
          {
            current_op->ids[device_index]=0xFF;
            current_op->data[device_index]=0x00000000;
          }
          current_op->num_devices--;
        }
      }
      if(current_op->num_devices==0x00)
        dyn_manager_delete_op_single_master(manager,i,op);
    }
  }
}

void dyn_manager_sync_op_change_address(TDynManager *manager,OP_HANDLE *op,unsigned short int address)
{
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
      manager->operations[i].sync_op[op->op_index[i]].address=address;
  }
}

void dyn_manager_sync_op_change_length(TDynManager *manager,OP_HANDLE *op,unsigned short int length)
{
  unsigned char i;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
      manager->operations[i].sync_op[op->op_index[i]].length=length;
  }
}

void dyn_manager_sync_op_change_data(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned char * const data[])
{
  TDynManagerSyncOp *current_op;
  unsigned char i,j,device_index;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
    {
      current_op=&manager->operations[i].sync_op[op->op_index[i]];
      for(j=0;j<num;j++)
      {
        if((device_index=dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[j]))!=0x00)
          current_op->data[device_index]=data[j];
      }
    }
  }
}

// bulk operation functions
OP_HANDLE *dyn_manager_bulk_op_new(TDynManager *manager,unsigned char mode,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  unsigned char i,master_index,current_op_index;
  TDynManagerBulkOp *current_op;
  OP_HANDLE *op_handle;

  if((op_handle=dyn_manager_get_free_op_handle(manager))==0x00000000)
    return 0x00000000;
  if(mode==DYN_MANAGER_READ || mode==DYN_MANAGER_WRITE)
    op_handle->op_type=(op_type_t)(DYN_MANAGER_BULK|mode);
  else
    return 0x00000000;
  for(i=0;i<num;i++)
  {
    if((master_index=dyn_manager_check_id(manager,ids[i]))==0xFF)
    {
      op_handle->op_type=no_op;
      return 0x00000000;
    }
    if(op_handle->op_index[master_index]==0xFF)
    {
      current_op_index=manager->operations[master_index].num_bulk_op;
      if(current_op_index<DYN_MANAGER_MAX_NUM_BULK_OP)
      {
        current_op=&manager->operations[master_index].bulk_op[current_op_index];
        if(dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[i])==0x00)
        {
          op_handle->op_index[master_index]=current_op_index;
          current_op->common.operation=(op_type_t)(DYN_MANAGER_BULK|mode);
          current_op->common.op_handle=op_handle;
          current_op->address[current_op->num_devices]=address[i];
          current_op->length[current_op->num_devices]=length[i];
          current_op->ids[current_op->num_devices]=ids[i];
          current_op->data[current_op->num_devices]=data[i];
          current_op->num_devices++;
          manager->operations[master_index].num_bulk_op++;
        }
      }
      else
      {
        op_handle->op_type=no_op;
        return 0x00000000;
      }
    }
    else
    {
      current_op_index=op_handle->op_index[master_index];
      current_op=&manager->operations[master_index].bulk_op[current_op_index];
      if(dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[i])==0x00)
      {
        current_op->address[current_op->num_devices]=address[i];
        current_op->length[current_op->num_devices]=length[i];
        current_op->ids[current_op->num_devices]=ids[i];
        current_op->data[current_op->num_devices]=data[i];
        current_op->num_devices++;
      }
    }
  }
  manager->num_ops++;
  return op_handle;
}

void dyn_manager_bulk_op_add_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length,unsigned char * const data[])
{
  TDynManagerBulkOp *current_op;
  unsigned char i,j;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
    {
      current_op=&manager->operations[i].bulk_op[op->op_index[i]];
      for(j=0;j<num;j++)
      {
        if(dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[j])==0x00)
        {
          if(current_op->num_devices<DYN_MANAGER_MAX_NUM_DEVICES)
          {
            current_op->ids[current_op->num_devices]=ids[i];
            current_op->address[current_op->num_devices]=address[i];
            current_op->length[current_op->num_devices]=length[i];
            current_op->data[current_op->num_devices]=data[i];
            current_op->num_devices++;
          }
        }
      }
    }
  }
}

void dyn_manager_bulk_op_remove_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids)
{
  TDynManagerBulkOp *current_op;
  unsigned char i,j,k,device_index;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
    {
      current_op=&manager->operations[i].bulk_op[op->op_index[i]];
      for(j=0;j<num;j++)
      {
        if((device_index=dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[j]))!=0x00)
        {
          if(device_index<(DYN_MANAGER_MAX_NUM_DEVICES-1))
          {
            for(k=device_index;k<current_op->num_devices;k++)
            {
              current_op->ids[k]=current_op->ids[k+1];
              current_op->address[k]=current_op->address[k+1];
              current_op->length[k]=current_op->length[k+1];
              current_op->data[k]=current_op->data[k+1];
            }
          }
          else
          {
            current_op->ids[device_index]=0xFF;
            current_op->address[device_index]=0x0000;
            current_op->length[device_index]=0x0000;
            current_op->data[device_index]=0x00000000;
          }
          current_op->num_devices--;
        }
      }
      if(current_op->num_devices==0x00)
        dyn_manager_delete_op_single_master(manager,i,op);
    }
  }
}

void dyn_manager_bulk_op_change_address(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned short int *address)
{
  TDynManagerBulkOp *current_op;
  unsigned char i,j,device_index;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
    {
      current_op=&manager->operations[i].bulk_op[op->op_index[i]];
      for(j=0;j<num;j++)
      {
        if((device_index=dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[j]))!=0x00)
          current_op->address[device_index]=address[j];
      }
    }
  }
}

void dyn_manager_bulk_op_change_length(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned short int *length)
{
  TDynManagerBulkOp *current_op;
  unsigned char i,j,device_index;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
    {
      current_op=&manager->operations[i].bulk_op[op->op_index[i]];
      for(j=0;j<num;j++)
      {
        if((device_index=dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[j]))!=0x00)
          current_op->length[device_index]=length[j];
      }
    }
  }
}

void dyn_manager_bulk_op_change_data(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned char * const data[])
{
  TDynManagerBulkOp *current_op;
  unsigned char i,j,device_index;

  for(i=0;i<manager->num_masters;i++)
  {
    if(op->op_index[i]!=0xFF)
    {
      current_op=&manager->operations[i].bulk_op[op->op_index[i]];
      for(j=0;j<num;j++)
      {
        if((device_index=dyn_manager_id_present(current_op->num_devices,current_op->ids,ids[j]))!=0x00)
          current_op->data[device_index]=data[j];
      }
    }
  }
}

void dyn_manager_loop(TDynManager *manager)
{
  unsigned char i,done,loop_done=manager->num_masters;
  dyn_manager_states_t states[DYN_MANAGER_MAX_NUM_MASTERS];
  unsigned char current_single_op[DYN_MANAGER_MAX_NUM_MASTERS];
  unsigned char current_sync_op[DYN_MANAGER_MAX_NUM_MASTERS];
  unsigned char current_bulk_op[DYN_MANAGER_MAX_NUM_MASTERS];
  TDynManagerSingleOp *single_op[DYN_MANAGER_MAX_NUM_MASTERS];
  TDynManagerSyncOp *sync_op[DYN_MANAGER_MAX_NUM_MASTERS];
  TDynManagerBulkOp *bulk_op[DYN_MANAGER_MAX_NUM_MASTERS];
  static unsigned char modules_period_count[DYN_MANAGER_MAX_NUM_MASTERS]={0};
  static unsigned char single_op_period_count[DYN_MANAGER_MAX_NUM_SINGLE_OP]={0};
  static unsigned char sync_op_period_count[DYN_MANAGER_MAX_NUM_SINGLE_OP]={0};
  static unsigned char bulk_op_period_count[DYN_MANAGER_MAX_NUM_SINGLE_OP]={0};

  // do scan
  if(manager->do_scan)
  {
    manager->do_scan=0x00;
//    dyn_manager_stop(manager);
    dyn_manager_scan(manager);
//    dyn_manager_start(manager);
    return;
  }

  // initialize the remaining operations
  for(i=0;i<manager->num_masters;i++)
  {
    states[i]=start_single_ops;
    current_single_op[i]=0;
    current_sync_op[i]=0;
    current_bulk_op[i]=0;
  }

  // execute the pre_process functions for all modules
  for(i=0;i<manager->num_modules;i++)
  {
    if(manager->modules[i]->enabled) 
    {
      modules_period_count[i]++;
      if(modules_period_count[i]==manager->modules[i]->period_count)
      {
        if(manager->modules[i]->pre_process!=0x00000000)
          manager->modules[i]->pre_process(manager->modules[i]->data);
      }
    }
  }

  // send all the commands
  do{
    for(i=0;i<manager->num_masters;i++)
    {
      switch(states[i])
      {
        case start_single_ops: if(current_single_op[i]>=manager->operations[i].num_single_op)
                                 states[i]=start_bulk_ops;
                               else
                               {
                                 single_op_period_count[current_single_op[i]]++;
                                 single_op[i]=&manager->operations[i].single_op[current_single_op[i]];
                                 if(single_op_period_count[current_single_op[i]]==single_op[i]->common.period_count)
                                 {
                                   if((single_op[i]->common.operation&DYN_MANAGER_MODE_MASK)==DYN_MANAGER_READ)
                                     dyn_master_start_read_table(manager->masters[i],single_op[i]->id,single_op[i]->address,single_op[i]->length,single_op[i]->data);
                                   else
                                     dyn_master_start_write_table(manager->masters[i],single_op[i]->id,single_op[i]->address,single_op[i]->length,single_op[i]->data);
                                   states[i]=wait_single_ops;
                                   single_op_period_count[current_single_op[i]]=0;
                                 }
                                 else
                                 {
                                   current_single_op[i]++;
                                   states[i]=start_single_ops;
                                 }
                               }
                               break;  
        case wait_single_ops: if((single_op[i]->common.operation&DYN_MANAGER_MODE_MASK)==DYN_MANAGER_READ)
                                done=dyn_master_is_read_table_done(manager->masters[i],single_op[i]->id,single_op[i]->address,single_op[i]->length,single_op[i]->data);
                              else
                                done=dyn_master_is_write_table_done(manager->masters[i],single_op[i]->id,single_op[i]->address,single_op[i]->length,single_op[i]->data);
                              if(done==DYN_BUSY)
                                states[i]=wait_single_ops;
                              else
                              {
                                if(single_op[i]->common.repetitions>0 && single_op[i]->common.repetitions!=0xFF)
                                  single_op[i]->common.repetitions--;
                                if(single_op[i]->common.repetitions==0)// remove operation from list 
                                  dyn_manager_delete_op_single_master(manager,i,single_op[i]->common.op_handle);
                                else
                                  current_single_op[i]++;
                                states[i]=start_single_ops;
                              }
                              break;
        case start_sync_ops: if(current_sync_op[i]>=manager->operations[i].num_sync_op)
                             {
                               states[i]=ops_done;
                               loop_done--;
                             }
                             else
                             {
                               sync_op_period_count[current_sync_op[i]]++;
                               sync_op[i]=&manager->operations[i].sync_op[current_sync_op[i]];
                               if(sync_op_period_count[current_sync_op[i]]==sync_op[i]->common.period_count)
                               {
                                 if((sync_op[i]->common.operation&DYN_MANAGER_MODE_MASK)==DYN_MANAGER_READ)
                                   dyn_master_start_sync_read(manager->masters[i],sync_op[i]->num_devices,sync_op[i]->ids,sync_op[i]->address,sync_op[i]->length,sync_op[i]->data);
                                 else
                                   dyn_master_start_sync_write(manager->masters[i],sync_op[i]->num_devices,sync_op[i]->ids,sync_op[i]->address,sync_op[i]->length,sync_op[i]->data);
                                 states[i]=wait_sync_ops;
                                 sync_op_period_count[current_sync_op[i]]=0;
                               }
                               else
                               {
                                 current_sync_op[i]++;
                                 states[i]=start_sync_ops;
                               }
                             }
                             break;
        case wait_sync_ops: if((sync_op[i]->common.operation&DYN_MANAGER_MODE_MASK)==DYN_MANAGER_READ)
                              done=dyn_master_is_sync_read_done(manager->masters[i],sync_op[i]->num_devices,sync_op[i]->ids,sync_op[i]->address,sync_op[i]->length,sync_op[i]->data);
                            else
                              done=dyn_master_is_sync_write_done(manager->masters[i],sync_op[i]->num_devices,sync_op[i]->ids,sync_op[i]->address,sync_op[i]->length,sync_op[i]->data);
                            if(done==DYN_BUSY)
                              states[i]=wait_sync_ops;
                            else
                            {
                              if(sync_op[i]->common.repetitions>0 && sync_op[i]->common.repetitions!=0xFF)
                                sync_op[i]->common.repetitions--;
                              if(sync_op[i]->common.repetitions==0)// remove operation from list 
                                dyn_manager_delete_op_single_master(manager,i,sync_op[i]->common.op_handle);
                              else
                                current_sync_op[i]++;
                              states[i]=start_sync_ops;
                            }
                            break;
        case start_bulk_ops: if(current_bulk_op[i]>=manager->operations[i].num_bulk_op)
                               states[i]=start_sync_ops;
                             else
                             {
                               bulk_op_period_count[current_bulk_op[i]]++;
                               bulk_op[i]=&manager->operations[i].bulk_op[current_bulk_op[i]];
                               if(bulk_op_period_count[current_bulk_op[i]]==bulk_op[i]->common.period_count)
                               {
                                 if((bulk_op[i]->common.operation&DYN_MANAGER_MODE_MASK)==DYN_MANAGER_READ)
                                   dyn_master_start_bulk_read(manager->masters[i],bulk_op[i]->num_devices,bulk_op[i]->ids,bulk_op[i]->address,bulk_op[i]->length,bulk_op[i]->data);
                                 else
                                   dyn_master_start_bulk_write(manager->masters[i],bulk_op[i]->num_devices,bulk_op[i]->ids,bulk_op[i]->address,bulk_op[i]->length,bulk_op[i]->data);
                                 states[i]=wait_bulk_ops;
                                 bulk_op_period_count[current_bulk_op[i]]=0;
                               }
                               else
                               {
                                 current_bulk_op[i]++;
                                 states[i]=start_bulk_ops;
                               }
                             }
                             break;
        case wait_bulk_ops: if((bulk_op[i]->common.operation&DYN_MANAGER_MODE_MASK)==DYN_MANAGER_READ)
                              done=dyn_master_is_bulk_read_done(manager->masters[i],bulk_op[i]->num_devices,bulk_op[i]->ids,bulk_op[i]->address,bulk_op[i]->length,bulk_op[i]->data);
                            else
                              done=dyn_master_is_bulk_write_done(manager->masters[i],bulk_op[i]->num_devices,bulk_op[i]->ids,bulk_op[i]->address,bulk_op[i]->length,bulk_op[i]->data);
                            if(done==DYN_BUSY)
                              states[i]=wait_bulk_ops;
                            else
                            {
                              if(bulk_op[i]->common.repetitions>0 && bulk_op[i]->common.repetitions!=0xFF)
                                bulk_op[i]->common.repetitions--;
                              if(bulk_op[i]->common.repetitions==0)// remove operation from list 
                                dyn_manager_delete_op_single_master(manager,i,bulk_op[i]->common.op_handle);
                              else
                                current_bulk_op[i]++;
                              states[i]=start_bulk_ops;
                            }
                            break;
        case ops_done: states[i]=ops_done;
                       break;
      }
    } 
  }while(loop_done>0);

  // execute the pre_process functions for all modules
  for(i=0;i<manager->num_modules;i++)
  {
    if(manager->modules[i]->enabled)
    {
      if(modules_period_count[i]==manager->modules[i]->period_count)
      {
        if(manager->modules[i]->post_process!=0x00000000)
          manager->modules[i]->post_process(manager->modules[i]->data);
        modules_period_count[i]=0;
      }
    }
  }
}
