#include <stdio.h>
#include <stdlib.h>
#include "dyn_devices.h"
#include "dyn_module.h"
#include "dyn_manager.h"
#include "motion_manager.h"
#include "action.h"
#include "dyn_servos.h"
#include "scheduler.h"

TDynamixelMaster dyn_master1;
TDynamixelMaster dyn_master2;
TDynManager manager;
TMemory memory;
TScheduler scheduler;

unsigned int SystemCoreClock=168000000;

// timer functions
unsigned long long int time_get_counts(void)
{
  static unsigned long long int count=0;
 
  count+=10000;
  return count;
}

unsigned int time_get_counts_per_us(void)
{
  return 10000;
}

// dynamixel master functions
unsigned char dyn_master_irq_receive_cb(void *dyn_master,unsigned char byte)
{

}

unsigned char dyn_master_dma_send_cb(void *dyn_master)
{

}

unsigned char dyn_master_dma_receive_cb(void *dyn_master)
{

}

void dummy_dyn_master_set_tx_mode(void)
{

}

void dummy_dyn_master_set_rx_mode(void)
{

}

void dyn_master_init(TDynamixelMaster *master,TComm *dev,TDynVersion version)
{
  /* assign communication functions */
  dev->irq_receive_cb=dyn_master_irq_receive_cb;
  dev->dma_send_cb=dyn_master_dma_send_cb;
  dev->dma_receive_cb=dyn_master_dma_receive_cb;
  master->comm_dev=dev;
  master->version=version;
  dev->data=master;
  /* initialize the internal callbacks */
  master->set_tx_mode=dummy_dyn_master_set_tx_mode;
  master->set_rx_mode=dummy_dyn_master_set_rx_mode;
  /* initialize internal variables */
  master->packet_ready=0x00;
  master->received_bytes=0x00;
  master->rx_timeout_ms=50;
  master->rx_no_answer=0x00;
  master->rx_num_packets=0x00;
  master->return_level=return_all;

  master->set_rx_mode();
}

void dyn_master_scan(TDynamixelMaster *master,unsigned char *num,unsigned char *ids)
{
  unsigned int i;

  if(master==&dyn_master1)
  {
    for(i=0;i<20;i++)
      ids[i]=i+1;
    (*num)=20;
  }
  else if(master==&dyn_master2)
  {
    for(i=0;i<4;i++)
      ids[i]=i+21;
    (*num)=4;
  }
}

unsigned char dyn_master_read_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  if(length==2)
    dyn_master_read_word(master,id,address,(unsigned short int *)data);
}

unsigned char dyn_master_read_word(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int *data)
{
  if(master==&dyn_master1)
  {
    if(id>0 && id<11)
    {
      if(address==0x0000)
        (*data)=SERVO_AX12A;
      else if(address==0x0006)
        (*data)=0x0000;
      else if(address==0x0008)
        (*data)=0x03FF;
      else if(address==0x0024)
        (*data)=0x01FF;
    } 
    if(id>10 && id<21)
    {
      if(address==0x0000)
        (*data)=SERVO_MX28;
      else if(address==0x0006)
        (*data)=0x0000;
      else if(address==0x0008)
        (*data)=0x0FFF;
      else if(address==0x0024)
        (*data)=0x07FF;
    } 
  }
  else if(master==&dyn_master2)
  {
    if(id>20 && id<25)
    {
      if(address==0x0000)
        (*data)=SERVO_XL320;
      else if(address==0x0006)
        (*data)=0x0000;
      else if(address==0x0008)
        (*data)=0x03FF;
      else if(address==0x0025)
        (*data)=0x01FF;
    } 
  }
}

unsigned char dyn_master_start_read_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
//  if(master==&dyn_master1)
//    printf("Master 1\n");
//  else
//    printf("Master 2\n");
//  printf("  read operation to servo %d at address %d with length %d\n",id,address,length);

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_read_table_done(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_write_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned short int i;

//  if(master==&dyn_master1)
//    printf("Master 1\n");
//  else
//    printf("Master 2\n");
//  printf("  write operation to servo %d at address %d with length %d\n",id,address,length);
//  printf("  data:\n");
//  for(i=0;i<length;i++)
//    printf("    0x%x\n",data[i]);

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_write_table_done(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_sync_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  unsigned short int i,j;

//  if(master==&dyn_master1)
//    printf("Master 1\n");
//  else
//    printf("Master 2\n");
//  printf("  sync write operation at address %d with length %d\n",address,length);
//  for(j=0;j<num;j++)
//  {
//    printf("  servo %d data:\n",ids[j]);
//    for(i=0;i<length;i++)
//      printf("    0x%x\n",data[j][i]);
//  }
  if(master==&dyn_master1 && address==ax_reg[cw_comp_slope].address && length==4)
  {
    for(i=0;i<num;i++)
      printf("%d,",data[i][2]+(data[i][3]<<8));
    printf("\n");
  }

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_sync_write_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_sync_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  unsigned short int i;

//  if(master==&dyn_master1)
//    printf("Master 1\n");
//  else
//    printf("Master 2\n");
//  printf("  sync read operation at address %d with length %d for servos:\n",address,length);
//  for(i=0;i<num;i++)
//    printf("  servo %d:\n",ids[i]);

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_sync_read_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_bulk_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  unsigned short int i,j;

//  if(master==&dyn_master1)
//    printf("Master 1\n");
//  else
//    printf("Master 2\n");
//  printf("  bulk write operation:\n");
//  for(j=0;j<num;j++)
//  {
//    printf("  servo %d at address %d with length %d and data:\n",ids[j],address[j],length[j]);
//    for(i=0;i<length[j];i++)
//      printf("    0x%x\n",data[j][i]);
//  }

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_bulk_write_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_bulk_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  unsigned short int i,j;

//  if(master==&dyn_master1)
//    printf("Master 1\n");
//  else
//    printf("Master 2\n");
//  printf("  bulk read operation:\n");
//  for(j=0;j<num;j++)
//    printf("  servo %d at address %d with length %d\n",ids[j],address[j],length[j]);

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_bulk_read_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  return DYN_SUCCESS;
}

int main(void)
{
  int i;
  TMotionManager mmanager;
  TActionMModule action;
  TTime dyn_master1_timer; 
  TComm dyn_master1_comm;
  TTime dyn_master2_timer; 
  TComm dyn_master2_comm;

  scheduler_init(&scheduler,1,100);

  // init masters
  time_init(&dyn_master1_timer,time_get_counts_per_us(),time_get_counts);
  comm_init(&dyn_master1_comm,0x01,&dyn_master1_timer);
  dyn_master_init(&dyn_master1,&dyn_master1_comm,DYN_VER1);
  time_init(&dyn_master2_timer,time_get_counts_per_us(),time_get_counts);
  comm_init(&dyn_master2_comm,0x01,&dyn_master2_timer);
  dyn_master_init(&dyn_master2,&dyn_master2_comm,DYN_VER2);

  dyn_manager_init(&manager,&memory,&scheduler,SCHED_CH1,0,64);
  dyn_manager_set_period(&manager,14);
  dyn_manager_add_master(&manager,&dyn_master1);
  dyn_manager_add_master(&manager,&dyn_master2);
  // init modules
  mmanager_init(&mmanager,&memory,16,128);
  action_init(&action,&memory,256);
  mmanager_add_module(&mmanager,action_get_module(&action));
  dyn_manager_add_module(&manager,mmanager_get_dyn_module(&mmanager));
  dyn_module_set_period(mmanager_get_dyn_module(&mmanager),1);
  dyn_module_enable(mmanager_get_dyn_module(&mmanager));
  dyn_manager_scan(&manager);

  /* enable all servos and assign them to the ACTION module */
  for(i=1;i<25;i++)
  {
    mmanager_set_module(&mmanager,i,MM_ACTION);
    mmanager_enable_servo(&mmanager,i);
  }
  /* load and start a page */
  if(action_load_page(&action,7))
    action_start_page(&action); 

  for(i=0;i<100;i++)
  {
//    printf("***** Iteration %d *****\n",i);
    dyn_manager_loop(&manager);
  }
}
