#include "stm32f1xx_hal.h"
#include "stm32_time.h"

#define ENABLE_LED_RX_GPIO_CLK      __HAL_RCC_GPIOC_CLK_ENABLE()
#define LED_RX_PIN                  GPIO_PIN_14                
#define LED_RX_GPIO_PORT            GPIOC                       

TIM_HandleTypeDef us_timer_Handle;
unsigned long long int counts;

void SysTick_Handler(void)
{
  HAL_IncTick();
}

void TIM7_IRQHandler(void)
{
  if(__HAL_TIM_GET_FLAG(&us_timer_Handle, TIM_FLAG_UPDATE) != RESET)
  {
    if(__HAL_TIM_GET_IT_SOURCE(&us_timer_Handle, TIM_IT_UPDATE) !=RESET)
    {
      __HAL_TIM_CLEAR_IT(&us_timer_Handle, TIM_IT_UPDATE);
      counts+=1;
    }
  }
}

void timer_init(void)
{
  us_timer_Handle.Instance=TIM7;
  us_timer_Handle.Init.Period=0xFFFF;
  us_timer_Handle.Init.Prescaler=0;
  us_timer_Handle.Init.ClockDivision=TIM_CLOCKDIVISION_DIV1;
  us_timer_Handle.Init.CounterMode=TIM_COUNTERMODE_UP;
  us_timer_Handle.Init.RepetitionCounter=0;
  HAL_TIM_Base_Init(&us_timer_Handle);
  __HAL_RCC_TIM7_CLK_ENABLE();
  __HAL_TIM_SET_COUNTER(&us_timer_Handle,0);

  // enable interrupts
  HAL_NVIC_SetPriority(TIM7_IRQn, 3, 3);
  HAL_NVIC_EnableIRQ(TIM7_IRQn);
  HAL_TIM_Base_Start_IT(&us_timer_Handle);

  counts=0;
}

unsigned long long int timer_get_counts(void)
{
  return (counts<<16)+__HAL_TIM_GET_COUNTER(&us_timer_Handle);
}

void gpio_init(void)
{
  GPIO_InitTypeDef GPIO_InitStructure;

  ENABLE_LED_RX_GPIO_CLK;

  GPIO_InitStructure.Pin = LED_RX_PIN;
  GPIO_InitStructure.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStructure.Speed = GPIO_SPEED_HIGH;
  GPIO_InitStructure.Pull = GPIO_NOPULL;
  HAL_GPIO_Init(LED_RX_GPIO_PORT, &GPIO_InitStructure);

  HAL_GPIO_WritePin(LED_RX_GPIO_PORT,LED_RX_PIN,GPIO_PIN_SET);
}

int main(void)
{
  uint32_t clocks_per_us;
  TTime time;
  uint8_t i;

  /* initialize the HAL module */
  HAL_Init();
  /* initialize the timer */
  timer_init();
  clocks_per_us=(SystemCoreClock/1000000)+1;
  time_init(&time,clocks_per_us,timer_get_counts);
  /* initialize the GPIO */
  gpio_init();

  while(1)                             /* main function does not return */
  {
    for(i=0;i<10;i++)
    {
      HAL_GPIO_WritePin(LED_RX_GPIO_PORT,LED_RX_PIN,GPIO_PIN_RESET);
      time_delay_ms(&time,1000);
      HAL_GPIO_WritePin(LED_RX_GPIO_PORT,LED_RX_PIN,GPIO_PIN_SET);
      time_delay_ms(&time,1000);
    }
    HAL_GPIO_WritePin(LED_RX_GPIO_PORT,LED_RX_PIN,GPIO_PIN_SET);
    time_set_timeout(&time,5000000);
    while(!time_is_timeout(&time));
    HAL_GPIO_WritePin(LED_RX_GPIO_PORT,LED_RX_PIN,GPIO_PIN_RESET);
  }
}

