#ifndef _FRAME_BUFFER_CONTROL_H
#define _FRAME_BUFFER_CONTROL_H

#ifdef __cplusplus
extern "C" {
#endif

#include "frame_buffer.h"
#include "patterns.h"
#include "memory.h"
#include "scheduler.h"

#ifndef FBC_MAX_PATTERN_FUNC
  #error "Please, specify the maximum number of pattern functions with the FBC_MAX_PATTERN_FUNC macro"
#endif

#ifndef FBC_MAX_DATA_PATTERN_LEN
  #error "Please, specify the maximum number of dtaa for each pattern with the FBC_MAX_DATA_PATTERN_LEN macro"
#endif

#ifndef FBC_MAX_COPY_FUNC
  #error "Please, specify the maximum number of copy functions with the FBC_MAX_COPY_FUNC macro"
#endif

#include "frame_buffer_control_registers.h"

typedef void (*copy_func_t)(void *data);

typedef struct TFBControl
{
  TFrameBuffer frame_buffer;
  TScheduler *scheduler;
  sched_channel_t sch_channel;
  unsigned short int period_ms;
  TMemory *memory;
  TMemModule mem_module;
  unsigned short int ram_base_address;
  unsigned short int eeprom_base_address;
  // pattern attributes
  unsigned char num_patterns;
  void (*pat_functions[FBC_MAX_PATTERN_FUNC])(TLEDArea *area,void *pattern_data,unsigned char buffer_id,unsigned short int period,struct TFBControl *control);
  TLEDArea pat_area[FBC_MAX_PATTERN_FUNC];
  unsigned char pat_buffer_id[FBC_MAX_PATTERN_FUNC];
  unsigned char pat_parameters[FBC_MAX_PATTERN_FUNC][FBC_MAX_DATA_PATTERN_LEN];
  // copy attributes
  unsigned char num_copy_functions;
  copy_func_t copy_functions[FBC_MAX_COPY_FUNC];
  void *copy_data[FBC_MAX_COPY_FUNC];
}TFBControl;

typedef void (*pattern_func_t)(TLEDArea *area,void *pattern_data,unsigned char buffer_id,unsigned short int period,TFBControl *control);

unsigned char frame_buffer_control_init(TFBControl *control,TMemory *memory,TScheduler *scheduler,sched_channel_t ch,unsigned short int ram_base_address,unsigned short int eeprom_base_address);
void frame_buffer_control_start(TFBControl *control);
void frame_buffer_control_stop(TFBControl *control);
void frame_buffer_control_set_period(TFBControl *control,unsigned short int period_ms);
unsigned short int frame_buffer_control_get_period(TFBControl *control);
// pattern functions
unsigned char frame_buffer_control_add_pattern(TFBControl *control,pattern_func_t function,TLEDArea *area,void *data,unsigned char data_length,unsigned char buffer_id);
void frame_buffer_control_remove_pattern(TFBControl *control,unsigned char index);
void frame_buffer_control_clear_patterns(TFBControl *control);
unsigned char frame_buffer_control_get_num_patterns(TFBControl *control);
unsigned char *frame_buffer_control_get_parameters(TFBControl *control, unsigned char index);
void frame_buffer_control_update(TFBControl *control);
// copy functions
unsigned char frame_buffer_control_add_copy_function(TFBControl *control,copy_func_t function,void *data);
void frame_buffer_control_remove_copy_function(TFBControl *control,unsigned char index);
void frame_buffer_control_clear_copy_functions(TFBControl *control);
unsigned char frame_buffer_control_get_num_copy_functions(TFBControl *control);

#ifdef __cplusplus
}
#endif

#endif
