#ifndef WAVEFORM_PATTERNS_H
#define WAVEFORM_PATTERNS_H

#include "patterns.h"
#include "frame_buffer_control.h"

#define WAVEFORM_GROUP               0x10

typedef enum {WF_SINE=0x01,WF_SQUARE=0x02,WF_RAMP=0x03,WF_TRIANGLE=0x04,WF_NONE=0x05} wf_id_t;

typedef struct
{
  unsigned short int phase;
}TWFSineData;

typedef struct
{
  unsigned short int t_on;
}TWFSquareData;

typedef struct
{
  unsigned short int dummy;
}TWFRampData;

typedef struct
{
  unsigned short int dummy;
}TWFTriangleData;

typedef union{
  TWFSineData sine_data;
  TWFSquareData square_data;
  TWFRampData ramp_data;
  TWFTriangleData trinagle_data;
}TWFData;

#pragma pack (push, 1)
typedef struct
{
  wf_id_t waveform_id;
  TPixelRGB max;
  TPixelRGB min;
  unsigned short int period;
  TWFData data;
  unsigned char active;
  // internal data (static)
  unsigned short int current_period;
}TWaveform;
#pragma pack (pop)

#define WF_ID_OFFSET              0
#define WF_MAX_R_OFFSET           1 
#define WF_MAX_G_OFFSET           2
#define WF_MAX_B_OFFSET           3
#define WF_MIN_R_OFFSET           4
#define WF_MIN_G_OFFSET           5
#define WF_MIN_B_OFFSET           6
#define WF_PERIOD_OFFSET          7
#define WF_DATA_OFFSET            9
#define WF_ACTIVE                 11
#define WF_CURRENT_PERIOD_OFFSET  12
void wf_general(TLEDArea *area,TWaveform *pattern_data,unsigned char buffer_id,unsigned short int period,TFBControl *control);

TPixelRGB wf_pixel(TWaveform *pattern_data);

#endif
