#include "frame_buffer.h"
#include "frame_buffer_control_registers.h"
#include "ram.h"

// private functions
void frame_buffer_compute_free_memory(TFrameBuffer *fb)
{
  fb->buffer_size=fb->num_rows*fb->num_pixels_per_row*FB_BYTES_PER_PIXEL;
  fb->free_mem=FB_MAX_BUFFER_LEN-fb->buffer_size*fb->num_buffers;
  fb->memory->data[fb->ram_base_address+FRAME_BUFFER_CONTROL_FREE_MEMORY_OFFSET]=fb->free_mem&0x00FF;
  fb->memory->data[fb->ram_base_address+FRAME_BUFFER_CONTROL_FREE_MEMORY_OFFSET+1]=(fb->free_mem&0xFF00)>>8;
}

// public functions
void frame_buffer_init(TFrameBuffer *fb,TMemory *memory,unsigned short int ram_base_address,unsigned short int eeprom_base_address)
{
  unsigned int i;

  /* initialize internal data */
  frame_buffer_set_num_rows(fb,DEFAULT_FRAME_BUFFER_CONTROL_NUM_ROWS);
  frame_buffer_set_pixels_per_row(fb,DEFAULT_FRAME_BUFFER_CONTROL_NUM_PIXELS);
  frame_buffer_set_num_buffers(fb,DEFAULT_FRAME_BUFFER_CONTROL_NUM_BUFFERS);
  fb->ram_base_address=ram_base_address;
  fb->eeprom_base_address=eeprom_base_address;
  fb->memory=memory;

  for(i=0;i<FB_MAX_BUFFER_LEN;i++)
    fb->pixel_buffer[i]=0x00;
}

void frame_buffer_set_num_rows(TFrameBuffer *fb,unsigned short int rows)
{
  fb->num_rows=rows;
  fb->memory->data[fb->eeprom_base_address+FRAME_BUFFER_CONTROL_NUM_ROWS_OFFSET]=rows&0x00FF;
  fb->memory->data[fb->eeprom_base_address+FRAME_BUFFER_CONTROL_NUM_ROWS_OFFSET+1]=(rows&0xFF00)>>8;
  frame_buffer_compute_free_memory(fb);
}

unsigned short int frame_buffer_get_num_rows(TFrameBuffer *fb)
{
  return fb->num_rows;
}

void frame_buffer_set_pixels_per_row(TFrameBuffer *fb,unsigned short int pixels)
{
  fb->num_pixels_per_row=pixels;
  fb->memory->data[fb->eeprom_base_address+FRAME_BUFFER_CONTROL_NUM_PIXELS_OFFSET]=pixels&0x00FF;
  fb->memory->data[fb->eeprom_base_address+FRAME_BUFFER_CONTROL_NUM_PIXELS_OFFSET+1]=(pixels&0xFF00)>>8;
  frame_buffer_compute_free_memory(fb);
}

unsigned short int frame_buffer_get_pixels_per_row(TFrameBuffer *fb)
{
  return fb->num_pixels_per_row;
}

void frame_buffer_set_num_buffers(TFrameBuffer *fb,unsigned char num)
{
  fb->num_buffers=num;
  fb->memory->data[fb->eeprom_base_address+FRAME_BUFFER_CONTROL_NUM_BUFFERS_OFFSET]=num&0x00FF;
  fb->memory->data[fb->eeprom_base_address+FRAME_BUFFER_CONTROL_NUM_BUFFERS_OFFSET+1]=(num&0xFF00)>>8;
  frame_buffer_compute_free_memory(fb);
}

unsigned char frame_buffer_get_num_buffers(TFrameBuffer *fb)
{
  return fb->num_buffers;
}

unsigned short int frame_buffer_get_free_memory(TFrameBuffer *fb)
{
  return fb->free_mem;
}

void frame_buffer_set_pixel(TFrameBuffer *fb,unsigned char buffer_id,unsigned char R,unsigned char G,unsigned char B,unsigned short int row,unsigned short int col)
{
  unsigned short int offset;

  if(buffer_id>=fb->num_buffers)
    return;
  else
  {
    offset=fb->num_rows*fb->num_pixels_per_row*FB_BYTES_PER_PIXEL*buffer_id;
    fb->pixel_buffer[offset+((row*fb->num_pixels_per_row+col)*FB_BYTES_PER_PIXEL)]=R;
    fb->pixel_buffer[offset+((row*fb->num_pixels_per_row+col)*FB_BYTES_PER_PIXEL)+1]=G;
    fb->pixel_buffer[offset+((row*fb->num_pixels_per_row+col)*FB_BYTES_PER_PIXEL)+2]=B;
  }
}

unsigned char *frame_buffer_get_buffer(TFrameBuffer *fb,unsigned char buffer_id)
{
  return &fb->pixel_buffer[fb->num_rows*fb->num_pixels_per_row*FB_BYTES_PER_PIXEL*buffer_id];
}
